#!/usr/bin/env python3
"""
Fig. 1 from:

Synchrony Generation in Recurrent Networks with Frequency-Dependent Synapses
The Journal of Neuroscience, 2000, Vol. 20 RC50

Implementation partially based on nest-2.0.0/examples/nest/tsodyks_shortterm_bursts.sli
by Moritz Helias, 2006.

Sebastian Schmitt, 2022
"""
import numpy as np

# set seed for reproducible figures
np.random.seed(5)

# for truncated normal
import scipy
from scipy import stats

import matplotlib.pyplot as plt

from brian2 import (
    NeuronGroup,
    Synapses,
    SpikeGeneratorGroup,
    SpikeMonitor,
    StateMonitor,
)
from brian2 import ms, mV
from brian2 import run, defaultclock


def truncated_normal(loc, scale, bounds, size):
    """Normal distribution truncated within bounds

    loc -- mean (“centre”) of the distribution
    scale -- standard deviation (spread or “width”) of the distribution
    bounds -- list of min and maximum
    size -- number of samples
    """
    bounds = np.array([bounds] * size)

    s = scipy.stats.truncnorm.rvs(
        (bounds[:, 0] - loc) / scale, (bounds[:, 1] - loc) / scale, loc=loc, scale=scale
    )

    return s


def get_population(name, N, tau_refrac):
    """Get population of neurons

    name -- name of population
    N -- number of neurons
    tau_refrac -- refractory period
    """

    neurons = NeuronGroup(
        N,
        """
        tau_mem : second
        tau_refrac : second
        v_reset : volt
        v_thresh : volt
        I_syn_ee_synapses : volt
        I_syn_ei_synapses : volt
        I_syn_ie_synapses : volt
        I_syn_ii_synapses : volt
        I_b : volt
        dv/dt = -v/tau_mem + (I_syn_ee_synapses +
                              I_syn_ei_synapses +
                              I_syn_ie_synapses +
                              I_syn_ii_synapses)/tau_mem
                           + I_b/tau_mem : volt (unless refractory)
        """,
        threshold="v>v_thresh",
        reset="v=v_reset",
        refractory=tau_refrac,
        method="exact",
        name=name,
    )

    v_thresh = 15 * mV
    v_reset = 13.5 * mV

    neurons.tau_mem = 30 * ms
    neurons.v_thresh = v_thresh
    neurons.v_reset = v_reset

    # paper gives range of 0.05 mV but population bursts are not visible with that value
    # -> increased to 1 mV range
    neurons.I_b = (
        np.random.uniform(v_thresh / mV - 0.5, v_thresh / mV + 0.5, size=N) * mV
    )

    return neurons


def get_synapses(name, source, target, tau_I, A, U, tau_rec, tau_facil=None):
    """Construct connections and retrieve synapses

    name -- name of synapses
    source -- source of connections
    target -- target of connections
    tau_I -- inactivation time constant
    A -- absolute synaptic strength
    U -- utilization of synaptic efficacy
    tau_rec -- recovery time constant
    tau_facil -- facilitation time constant (optional)
    """

    synapses_eqs = """
    A : volt
    U : 1
    tau_I : second
    tau_rec : second

    dx/dt =  z/tau_rec : 1 (clock-driven) # recovered
    dy/dt = -y/tau_I   : 1 (clock-driven) # active
    z = 1 - x - y      : 1                # inactive
    I_syn_{}_post = A*y : volt (summed)
    """.format(
        name
    )

    if tau_facil:
        synapses_eqs += """
        du/dt = -u/tau_facil : 1 (clock-driven)
        tau_facil : second
        """

        synapses_action = """
        u += U*(1-u)
        y += u*x # important: update y first
        x += -u*x
        """
    else:
        synapses_action = """
        y += U*x # important: update y first
        x += -U*x
        """

    synapses = Synapses(
        source,
        target,
        model=synapses_eqs,
        on_pre=synapses_action,
        method="exact",
        name=name,
    )
    synapses.connect(p=0.1)

    N_syn = len(synapses)

    synapses.tau_I = tau_I

    A_min = min(0.2 * A, 2 * A)
    A_max = max(0.2 * A, 2 * A)
    synapses.A = (
        truncated_normal(
            A / mV, 0.5 * abs(A / mV), [A_min / mV, A_max / mV], size=N_syn
        ) * mV
    )
    assert not any(synapses.A < A_min)
    assert not any(synapses.A > A_max)

    U_mean, U_min, U_max = U
    synapses.U = truncated_normal(U_mean, 0.5 * U_mean, [U_min, U_max], size=N_syn)
    assert not any(synapses.U <= U_min)
    assert not any(synapses.U > U_max)

    tau_min = 5
    synapses.tau_rec = (
        truncated_normal(
            tau_rec / ms, 0.5 * tau_rec / ms, [tau_min, np.inf], size=N_syn
        ) * ms
    )
    assert not any(synapses.tau_rec / ms <= tau_min)

    if tau_facil:
        synapses.tau_facil = (
            truncated_normal(
                tau_facil / ms, 0.5 * tau_facil / ms, [tau_min, np.inf], size=N_syn
            ) * ms
        )
        assert not any(synapses.tau_facil / ms <= tau_min)

    # start fully recovered
    synapses.x = 1

    return synapses


# configure neuron populations
exc_neurons = get_population("exc_neurons", N=400, tau_refrac=3 * ms)
inh_neurons = get_population("inh_neurons", N=100, tau_refrac=2 * ms)

# configure synapses
ee_synapses = get_synapses(
    "ee_synapses",
    exc_neurons,
    exc_neurons,
    tau_I=3 * ms,
    A=1.8 * mV,
    U=[0.5, 0.1, 0.9],
    tau_rec=800 * ms,
)
ei_synapses = get_synapses(
    "ei_synapses",
    exc_neurons,
    inh_neurons,
    tau_I=3 * ms,
    A=7.2 * mV,
    U=[0.04, 0.001, 0.07],
    tau_rec=100 * ms,
    tau_facil=1000 * ms,
)
ie_synapses = get_synapses(
    "ie_synapses",
    inh_neurons,
    exc_neurons,
    tau_I=3 * ms,
    A=-5.4 * mV,
    U=[0.5, 0.1, 0.9],
    tau_rec=800 * ms,
)
ii_synapses = get_synapses(
    "ii_synapses",
    inh_neurons,
    inh_neurons,
    tau_I=3 * ms,
    A=-7.2 * mV,
    U=[0.04, 0.001, 0.07],
    tau_rec=100 * ms,
    tau_facil=1000 * ms,
)

# run for burnin time to settle network activity
defaultclock.dt = 1 * ms
burnin = 900
run(burnin * ms)

# record from now on
spike_monitor_exc = SpikeMonitor(exc_neurons)
spike_monitor_inh = SpikeMonitor(inh_neurons)
state_monitor_ee = StateMonitor(ee_synapses, ["x"], record=True)

duration = 4200
run(duration * ms, report="text")

# plots
fig, axes = plt.subplots(3, figsize=(6, 8), sharex=True)

# raster plot
axes[0].plot(spike_monitor_exc.t / ms, spike_monitor_exc.i, ".k", ms=1)
axes[0].plot(spike_monitor_inh.t / ms, spike_monitor_inh.i + len(exc_neurons), ".k", ms=1)
axes[0].set_ylabel("Neuron No.")
axes[0].set_ylim(0, len(exc_neurons) + len(inh_neurons))

# network activity
net_activity = np.histogram(
    np.concatenate(
        list(spike_monitor_exc.spike_trains().values())
        + list(spike_monitor_inh.spike_trains().values())
    ) / ms,
    bins=np.arange(burnin, duration + burnin, 1))[0] / (len(exc_neurons) + len(inh_neurons))
axes[1].plot(np.arange(0, len(net_activity)) + burnin, net_activity, "k")
net_activity_min = 0
net_activity_max = 0.2
axes[1].set_ylim(net_activity_min, net_activity_max)
axes[1].set_ylabel("Net activity")

# network activity inset
axins = axes[1].inset_axes([0.05, 0.35, 0.2, 0.6])
axins.plot(np.arange(0, len(net_activity)) + burnin, net_activity, "k")
inset_min = 1220
inset_max = 1260
axins.set_xlim(inset_min + burnin, inset_max + burnin)
axins.set_ylim(net_activity_min, net_activity_max)
axins.set_xticks([inset_min + burnin, inset_max + burnin])
axins.set_xticklabels([inset_min, inset_max])
axins.set_yticks([])

# recovered synaptic partition
axes[2].plot(
    state_monitor_ee.t / ms, np.mean(state_monitor_ee.x, axis=0), "k", label="x"
)
axes[2].set_ylim(0.2, 0.6)
axes[2].set_xlabel("Time (msec)")
axes[2].set_ylabel("Recov excit")
axes[2].set_xlim(burnin, duration + burnin)
xtickstep = 1000
axes[2].set_xticks(np.arange(burnin, duration + burnin, xtickstep))
axes[2].set_xticklabels(map(str, range(0, duration, xtickstep)))

axes[0].xaxis.set_tick_params(which="both", labelbottom=True)
axes[1].xaxis.set_tick_params(which="both", labelbottom=True)

plt.show()



"""
Classes used to specify the type of a function, variable or common
sub-expression.
"""

import collections
import functools
import numbers
from collections.abc import Mapping

import numpy as np

from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    Dimension,
    Quantity,
    fail_for_dimension_mismatch,
    get_unit,
    get_unit_for_display,
)
from brian2.utils.caching import CacheKey
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers, word_substitute

from .base import device_override, weakproxy_with_fallback
from .preferences import prefs

__all__ = [
    "Variable",
    "Constant",
    "ArrayVariable",
    "DynamicArrayVariable",
    "Subexpression",
    "AuxiliaryVariable",
    "VariableView",
    "Variables",
    "LinkedVariable",
    "linked_var",
]


logger = get_logger(__name__)


def get_dtype(obj):
    """
    Helper function to return the `numpy.dtype` of an arbitrary object.

    Parameters
    ----------
    obj : object
        Any object (but typically some kind of number or array).

    Returns
    -------
    dtype : `numpy.dtype`
        The type of the given object.
    """
    if hasattr(obj, "dtype"):
        return obj.dtype
    else:
        return np.dtype(type(obj))


def get_dtype_str(val):
    """
    Returns canonical string representation of the dtype of a value or dtype

    Returns
    -------

    dtype_str : str
        The numpy dtype name
    """
    if isinstance(val, np.dtype):
        return val.name
    if isinstance(val, type):
        return get_dtype_str(val())

    is_bool = val is True or val is False or val is np.True_ or val is np.False_
    if is_bool:
        return "bool"
    if hasattr(val, "dtype"):
        return get_dtype_str(val.dtype)
    if isinstance(val, numbers.Number):
        return get_dtype_str(np.array(val).dtype)

    return f"unknown[{str(val)}, {val.__class__.__name__}]"


def variables_by_owner(variables, owner):
    owner_name = getattr(owner, "name", None)
    return {
        varname: var
        for varname, var in variables.items()
        if getattr(var.owner, "name", None) is owner_name
    }


class Variable(CacheKey):
    r"""
    An object providing information about model variables (including implicit
    variables such as ``t`` or ``xi``). This class should never be
    instantiated outside of testing code, use one of its subclasses instead.

    Parameters
    ----------
    name : 'str'
        The name of the variable. Note that this refers to the *original*
        name in the owning group. The same variable may be known under other
        names in other groups (e.g. the variable ``v`` of a `NeuronGroup` is
        known as ``v_post`` in a `Synapse` connecting to the group).
    dimensions : `Dimension`, optional
        The physical dimensions of the variable.
    owner : `Nameable`, optional
        The object that "owns" this variable, e.g. the `NeuronGroup` or
        `Synapses` object that declares the variable in its model equations.
        Defaults to ``None`` (the value used for `Variable` objects without an
        owner, e.g. external `Constant`\ s).
    dtype : `dtype`, optional
        The dtype used for storing the variable. Defaults to the preference
        `core.default_scalar.dtype`.
    scalar : bool, optional
        Whether the variable is a scalar value (``True``) or vector-valued, e.g.
        defined for every neuron (``False``). Defaults to ``False``.
    constant: bool, optional
        Whether the value of this variable can change during a run. Defaults
        to ``False``.
    read_only : bool, optional
        Whether this is a read-only variable, i.e. a variable that is set
        internally and cannot be changed by the user (this is used for example
        for the variable ``N``, the number of neurons in a group). Defaults
        to ``False``.
    array : bool, optional
        Whether this variable is an array. Allows for simpler check than testing
        ``isinstance(var, ArrayVariable)``. Defaults to ``False``.
    """

    _cache_irrelevant_attributes = {"owner"}

    def __init__(
        self,
        name,
        dimensions=DIMENSIONLESS,
        owner=None,
        dtype=None,
        scalar=False,
        constant=False,
        read_only=False,
        dynamic=False,
        array=False,
    ):
        assert isinstance(dimensions, Dimension)

        #: The variable's dimensions.
        self.dim = dimensions

        #: The variable's name.
        self.name = name

        #: The `Group` to which this variable belongs.
        self.owner = weakproxy_with_fallback(owner) if owner is not None else None

        #: The dtype used for storing the variable.
        self.dtype = dtype
        if dtype is None:
            self.dtype = prefs.core.default_float_dtype

        if self.is_boolean:
            if dimensions is not DIMENSIONLESS:
                raise ValueError("Boolean variables can only be dimensionless")

        #: Whether the variable is a scalar
        self.scalar = scalar

        #: Whether the variable is constant during a run
        self.constant = constant

        #: Whether the variable is read-only
        self.read_only = read_only

        #: Whether the variable is dynamically sized (only for non-scalars)
        self.dynamic = dynamic

        #: Whether the variable is an array
        self.array = array

    def __getstate__(self):
        state = self.__dict__.copy()
        state["owner"] = state["owner"].__repr__.__self__  # replace proxy
        return state

    def __setstate__(self, state):
        state["owner"] = weakproxy_with_fallback(state["owner"])
        self.__dict__ = state

    @property
    def is_boolean(self):
        return np.issubdtype(self.dtype, np.bool_)

    @property
    def is_integer(self):
        return np.issubdtype(self.dtype, np.signedinteger)

    @property
    def dtype_str(self):
        """
        String representation of the numpy dtype
        """
        return get_dtype_str(self)

    @property
    def unit(self):
        """
        The `Unit` of this variable
        """
        return get_unit(self.dim)

    def get_value(self):
        """
        Return the value associated with the variable (without units). This
        is the way variables are accessed in generated code.
        """
        raise TypeError(f"Cannot get value for variable {self}")

    def set_value(self, value):
        """
        Set the value associated with the variable.
        """
        raise TypeError(f"Cannot set value for variable {self}")

    def get_value_with_unit(self):
        """
        Return the value associated with the variable (with units).
        """
        return Quantity(self.get_value(), self.dim)

    def get_addressable_value(self, name, group):
        """
        Get the value (without units) of this variable in a form that can be
        indexed in the context of a group. For example, if a
        postsynaptic variable ``x`` is accessed in a synapse ``S`` as
        ``S.x_post``, the synaptic indexing scheme can be used.

        Parameters
        ----------
        name : str
            The name of the variable
        group : `Group`
            The group providing the context for the indexing. Note that this
            `group` is not necessarily the same as `Variable.owner`: a variable
            owned by a `NeuronGroup` can be indexed in a different way if
            accessed via a `Synapses` object.

        Returns
        -------
        variable : object
            The variable in an indexable form (without units).
        """
        return self.get_value()

    def get_addressable_value_with_unit(self, name, group):
        """
        Get the value (with units) of this variable in a form that can be
        indexed in the context of a group. For example, if a postsynaptic
        variable ``x`` is accessed in a synapse ``S`` as ``S.x_post``, the
        synaptic indexing scheme can be used.

        Parameters
        ----------
        name : str
            The name of the variable
        group : `Group`
            The group providing the context for the indexing. Note that this
            `group` is not necessarily the same as `Variable.owner`: a variable
            owned by a `NeuronGroup` can be indexed in a different way if
            accessed via a `Synapses` object.

        Returns
        -------
        variable : object
            The variable in an indexable form (with units).
        """
        return self.get_value_with_unit()

    def get_len(self):
        """
        Get the length of the value associated with the variable or ``0`` for
        a scalar variable.
        """
        if self.scalar:
            return 0
        else:
            return len(self.get_value())

    def __len__(self):
        return self.get_len()

    def __repr__(self):
        description = (
            "<{classname}(dimensions={dimensions}, "
            " dtype={dtype}, scalar={scalar}, constant={constant},"
            " read_only={read_only})>"
        )
        return description.format(
            classname=self.__class__.__name__,
            dimensions=repr(self.dim),
            dtype=getattr(self.dtype, "__name__", repr(self.dtype)),
            scalar=repr(self.scalar),
            constant=repr(self.constant),
            read_only=repr(self.read_only),
        )


# ------------------------------------------------------------------------------
# Concrete classes derived from `Variable` -- these are the only ones ever
# instantiated.
# ------------------------------------------------------------------------------


class Constant(Variable):
    """
    A scalar constant (e.g. the number of neurons ``N``). Information such as
    the dtype or whether this variable is a boolean are directly derived from
    the `value`. Most of the time `Variables.add_constant` should be used
    instead of instantiating this class directly.

    Parameters
    ----------
    name : str
        The name of the variable
    dimensions : `Dimension`, optional
        The physical dimensions of the variable. Note that the variable itself
        (as referenced by value) should never have units attached.
    value: reference to the variable value
        The value of the constant.
    owner : `Nameable`, optional
        The object that "owns" this variable, for constants that belong to a
        specific group, e.g. the ``N`` constant for a `NeuronGroup`. External
        constants will have ``None`` (the default value).
    """

    def __init__(self, name, value, dimensions=DIMENSIONLESS, owner=None):
        # Determine the type of the value
        is_bool = (
            value is True or value is False or value is np.True_ or value is np.False_
        )

        if is_bool:
            dtype = bool
        else:
            dtype = get_dtype(value)

        # Use standard Python types if possible for numpy scalars
        if getattr(value, "shape", None) == () and hasattr(value, "dtype"):
            numpy_type = value.dtype
            if np.can_cast(numpy_type, int):
                value = int(value)
            elif np.can_cast(numpy_type, float):
                value = float(value)
            elif np.can_cast(numpy_type, complex):
                value = complex(value)
            elif value is np.True_:
                value = True
            elif value is np.False_:
                value = False

        #: The constant's value
        self.value = value

        super().__init__(
            dimensions=dimensions,
            name=name,
            owner=owner,
            dtype=dtype,
            scalar=True,
            constant=True,
            read_only=True,
        )

    def get_value(self):
        return self.value

    def item(self):
        return self.value


class AuxiliaryVariable(Variable):
    """
    Variable description for an auxiliary variable (most likely one that is
    added automatically to abstract code, e.g. ``_cond`` for a threshold
    condition), specifying its type and unit for code generation. Most of the
    time `Variables.add_auxiliary_variable` should be used instead of
    instantiating this class directly.

    Parameters
    ----------
    name : str
        The name of the variable
    dimensions : `Dimension`, optional
        The physical dimensions of the variable.
    dtype : `dtype`, optional
        The dtype used for storing the variable. If none is given, defaults
        to `core.default_float_dtype`.
    scalar : bool, optional
        Whether the variable is a scalar value (``True``) or vector-valued, e.g.
        defined for every neuron (``False``). Defaults to ``False``.
    """

    def __init__(self, name, dimensions=DIMENSIONLESS, dtype=None, scalar=False):
        super().__init__(dimensions=dimensions, name=name, dtype=dtype, scalar=scalar)

    def get_value(self):
        raise TypeError(
            f"Cannot get the value for an auxiliary variable ({self.name})."
        )


class ArrayVariable(Variable):
    """
    An object providing information about a model variable stored in an array
    (for example, all state variables). Most of the time `Variables.add_array`
    should be used instead of instantiating this class directly.

    Parameters
    ----------
    name : 'str'
        The name of the variable. Note that this refers to the *original*
        name in the owning group. The same variable may be known under other
        names in other groups (e.g. the variable ``v`` of a `NeuronGroup` is
        known as ``v_post`` in a `Synapse` connecting to the group).
    dimensions : `Dimension`, optional
        The physical dimensions of the variable
    owner : `Nameable`
        The object that "owns" this variable, e.g. the `NeuronGroup` or
        `Synapses` object that declares the variable in its model equations.
    size : int
        The size of the array
    device : `Device`
        The device responsible for the memory access.
    dtype : `dtype`, optional
        The dtype used for storing the variable. If none is given, defaults
        to `core.default_float_dtype`.
    constant : bool, optional
        Whether the variable's value is constant during a run.
        Defaults to ``False``.
    scalar : bool, optional
        Whether this array is a 1-element array that should be treated like a
        scalar (e.g. for a single delay value across synapses). Defaults to
        ``False``.
    read_only : bool, optional
        Whether this is a read-only variable, i.e. a variable that is set
        internally and cannot be changed by the user. Defaults
        to ``False``.
    unique : bool, optional
        Whether the values in this array are all unique. This information is
        only important for variables used as indices and does not have to
        reflect the actual contents of the array but only the possibility of
        non-uniqueness (e.g. synaptic indices are always unique but the
        corresponding pre- and post-synaptic indices are not). Defaults to
        ``False``.
    """

    def __init__(
        self,
        name,
        owner,
        size,
        device,
        dimensions=DIMENSIONLESS,
        dtype=None,
        constant=False,
        scalar=False,
        read_only=False,
        dynamic=False,
        unique=False,
    ):
        super().__init__(
            dimensions=dimensions,
            name=name,
            owner=owner,
            dtype=dtype,
            scalar=scalar,
            constant=constant,
            read_only=read_only,
            dynamic=dynamic,
            array=True,
        )

        #: Wether all values in this arrays are necessarily unique (only
        #: relevant for index variables).
        self.unique = unique

        #: The `Device` responsible for memory access.
        self.device = device

        #: The size of this variable.
        self.size = size

        if scalar and size != 1:
            raise ValueError(f"Scalar variables need to have size 1, not size {size}.")

        #: Another variable, on which the write is conditioned (e.g. a variable
        #: denoting the absence of refractoriness)
        self.conditional_write = None

    def set_conditional_write(self, var):
        if not var.is_boolean:
            raise TypeError(
                "A variable can only be conditionally writeable "
                f"depending on a boolean variable, '{var.name}' is not "
                "boolean."
            )
        self.conditional_write = var

    def get_value(self):
        return self.device.get_value(self)

    def item(self):
        if self.size == 1:
            return self.get_value().item()
        else:
            raise ValueError("can only convert an array of size 1 to a Python scalar")

    def set_value(self, value):
        self.device.fill_with_array(self, value)

    def get_len(self):
        return self.size

    def get_addressable_value(self, name, group):
        return VariableView(name=name, variable=self, group=group, dimensions=None)

    def get_addressable_value_with_unit(self, name, group):
        return VariableView(name=name, variable=self, group=group, dimensions=self.dim)


class DynamicArrayVariable(ArrayVariable):
    """
    An object providing information about a model variable stored in a dynamic
    array (used in `Synapses`). Most of the time `Variables.add_dynamic_array`
    should be used instead of instantiating this class directly.

    Parameters
    ----------
    name : 'str'
        The name of the variable. Note that this refers to the *original*
        name in the owning group. The same variable may be known under other
        names in other groups (e.g. the variable ``v`` of a `NeuronGroup` is
        known as ``v_post`` in a `Synapse` connecting to the group).
    dimensions : `Dimension`, optional
        The physical dimensions of the variable.
    owner : `Nameable`
        The object that "owns" this variable, e.g. the `NeuronGroup` or
        `Synapses` object that declares the variable in its model equations.
    size : int or tuple of int
        The (initial) size of the variable.
    device : `Device`
        The device responsible for the memory access.
    dtype : `dtype`, optional
        The dtype used for storing the variable. If none is given, defaults
        to `core.default_float_dtype`.
    constant : bool, optional
        Whether the variable's value is constant during a run.
        Defaults to ``False``.
    needs_reference_update : bool, optional
        Whether the code objects need a new reference to the underlying data at
        every time step. This should be set if the size of the array can be
        changed by other code objects. Defaults to ``False``.
    scalar : bool, optional
        Whether this array is a 1-element array that should be treated like a
        scalar (e.g. for a single delay value across synapses). Defaults to
        ``False``.
    read_only : bool, optional
        Whether this is a read-only variable, i.e. a variable that is set
        internally and cannot be changed by the user. Defaults
        to ``False``.
    unique : bool, optional
        Whether the values in this array are all unique. This information is
        only important for variables used as indices and does not have to
        reflect the actual contents of the array but only the possibility of
        non-uniqueness (e.g. synaptic indices are always unique but the
        corresponding pre- and post-synaptic indices are not). Defaults to
        ``False``.
    """

    # The size of a dynamic variable can of course change and changes in
    # size should not invalidate the cache
    _cache_irrelevant_attributes = ArrayVariable._cache_irrelevant_attributes | {"size"}

    def __init__(
        self,
        name,
        owner,
        size,
        device,
        dimensions=DIMENSIONLESS,
        dtype=None,
        constant=False,
        needs_reference_update=False,
        resize_along_first=False,
        scalar=False,
        read_only=False,
        unique=False,
    ):
        if isinstance(size, int):
            ndim = 1
        else:
            ndim = len(size)

        #: The number of dimensions
        self.ndim = ndim

        if constant and needs_reference_update:
            raise ValueError("A variable cannot be constant and need reference updates")
        #: Whether this variable needs an update of the reference to the
        #: underlying data whenever it is passed to a code object
        self.needs_reference_update = needs_reference_update

        #: Whether this array will be only resized along the first dimension
        self.resize_along_first = resize_along_first

        super().__init__(
            dimensions=dimensions,
            owner=owner,
            name=name,
            size=size,
            device=device,
            constant=constant,
            dtype=dtype,
            scalar=scalar,
            dynamic=True,
            read_only=read_only,
            unique=unique,
        )

    @property
    def dimensions(self):
        logger.warn(
            "The DynamicArrayVariable.dimensions attribute is "
            "deprecated, use .ndim instead",
            "deprecated_dimensions",
            once=True,
        )
        return self.ndim

    def resize(self, new_size):
        """
        Resize the dynamic array. Calls `self.device.resize` to do the
        actual resizing.

        Parameters
        ----------
        new_size : int or tuple of int
            The new size.
        """
        if self.resize_along_first:
            self.device.resize_along_first(self, new_size)
        else:
            self.device.resize(self, new_size)

        self.size = new_size


class Subexpression(Variable):
    """
    An object providing information about a named subexpression in a model.
    Most of the time `Variables.add_subexpression` should be used instead of
    instantiating this class directly.

    Parameters
    ----------
    name : str
        The name of the subexpression.
    dimensions : `Dimension`, optional
        The physical dimensions of the subexpression.
    owner : `Group`
        The group to which the expression refers.
    expr : str
        The subexpression itself.
    device : `Device`
        The device responsible for the memory access.
    dtype : `dtype`, optional
        The dtype used for the expression. Defaults to
        `core.default_float_dtype`.
    scalar: bool, optional
        Whether this is an expression only referring to scalar variables.
        Defaults to ``False``
    """

    def __init__(
        self,
        name,
        owner,
        expr,
        device,
        dimensions=DIMENSIONLESS,
        dtype=None,
        scalar=False,
    ):
        super().__init__(
            dimensions=dimensions,
            owner=owner,
            name=name,
            dtype=dtype,
            scalar=scalar,
            constant=False,
            read_only=True,
        )

        #: The `Device` responsible for memory access
        self.device = device

        #: The expression defining the subexpression
        self.expr = expr.strip()

        #: The identifiers used in the expression
        self.identifiers = get_identifiers(expr)

    def get_addressable_value(self, name, group):
        return VariableView(
            name=name, variable=self, group=group, dimensions=DIMENSIONLESS
        )

    def get_addressable_value_with_unit(self, name, group):
        return VariableView(name=name, variable=self, group=group, dimensions=self.dim)

    def __contains__(self, var):
        return var in self.identifiers

    def __repr__(self):
        description = (
            "<{classname}(name={name}, dimensions={dimensions}, dtype={dtype}, "
            "expr={expr}, owner=<{owner}>)>"
        )
        return description.format(
            classname=self.__class__.__name__,
            name=repr(self.name),
            dimensions=repr(self.dim),
            dtype=repr(self.dtype),
            expr=repr(self.expr),
            owner=self.owner.name,
        )


# ------------------------------------------------------------------------------
# Classes providing views on variables and storing variables information
# ------------------------------------------------------------------------------
class LinkedVariable:
    """
    A simple helper class to make linking variables explicit. Users should use
    `linked_var` instead.

    Parameters
    ----------
    group : `Group`
        The group through which the `variable` is accessed (not necessarily the
        same as ``variable.owner``.
    name : str
        The name of `variable` in `group` (not necessarily the same as
         ``variable.name``).
    variable : `Variable`
        The variable that should be linked.
    index : str or `ndarray`, optional
        An indexing array (or the name of a state variable), providing a mapping
        from the entries in the link source to the link target.
    """

    def __init__(self, group, name, variable, index=None):
        if isinstance(variable, DynamicArrayVariable):
            raise NotImplementedError(
                f"Linking to variable {variable.name} is "
                "not supported, can only link to "
                "state variables of fixed size."
            )
        self.group = group
        self.name = name
        self.variable = variable
        self.index = index


def linked_var(group_or_variable, name=None, index=None):
    """
    Represents a link target for setting a linked variable.

    Parameters
    ----------
    group_or_variable : `NeuronGroup` or `VariableView`
        Either a reference to the target `NeuronGroup` (e.g. ``G``) or a direct
        reference to a `VariableView` object (e.g. ``G.v``). In case only the
        group is specified, `name` has to be specified as well.
    name : str, optional
        The name of the target variable, necessary if `group_or_variable` is a
        `NeuronGroup`.
    index : str or `ndarray`, optional
        An indexing array (or the name of a state variable), providing a mapping
        from the entries in the link source to the link target.

    Examples
    --------
    >>> from brian2 import *
    >>> G1 = NeuronGroup(10, 'dv/dt = -v / (10*ms) : volt')
    >>> G2 = NeuronGroup(10, 'v : volt (linked)')
    >>> G2.v = linked_var(G1, 'v')
    >>> G2.v = linked_var(G1.v)  # equivalent
    """
    if isinstance(group_or_variable, VariableView):
        if name is not None:
            raise ValueError(
                "Cannot give a variable and a variable name at the same time."
            )
        return LinkedVariable(
            group_or_variable.group,
            group_or_variable.name,
            group_or_variable.variable,
            index=index,
        )
    elif name is None:
        raise ValueError("Need to provide a variable name")
    else:
        return LinkedVariable(
            group_or_variable, name, group_or_variable.variables[name], index=index
        )


class VariableView:
    """
    A view on a variable that allows to treat it as an numpy array while
    allowing special indexing (e.g. with strings) in the context of a `Group`.

    Parameters
    ----------
    name : str
        The name of the variable (not necessarily the same as ``variable.name``).
    variable : `Variable`
        The variable description.
    group : `Group`
        The group through which the variable is accessed (not necessarily the
        same as `variable.owner`).
    dimensions : `Dimension`, optional
        The physical dimensions to be used for the variable, should be `None`
        when a variable is accessed without units (e.g. when accessing
        ``G.var_``).
    """

    __array_priority__ = 10

    def __init__(self, name, variable, group, dimensions=None):
        self.name = name
        self.variable = variable
        self.index_var_name = group.variables.indices[name]
        if self.index_var_name in ("_idx", "0"):
            self.index_var = self.index_var_name
        else:
            self.index_var = group.variables[self.index_var_name]

        if isinstance(variable, Subexpression):
            # For subexpressions, we *always* have to go via codegen to get
            # their value -- since we cannot do this without the group, we
            # hold a strong reference
            self.group = group
        else:
            # For state variable arrays, we can do most access without the full
            # group, using the indexing reference below. We therefore only keep
            # a weak reference to the group.
            self.group = weakproxy_with_fallback(group)
        self.group_name = group.name
        # We keep a strong reference to the `Indexing` object so that basic
        # indexing is still possible, even if the group no longer exists
        self.indexing = self.group._indices
        self.dim = dimensions

    @property
    def unit(self):
        """
        The `Unit` of this variable
        """
        return get_unit(self.dim)

    def get_item(self, item, level=0, namespace=None):
        """
        Get the value of this variable. Called by `__getitem__`.

        Parameters
        ----------
        item : slice, `ndarray` or string
            The index for the setting operation
        level : int, optional
            How much farther to go up in the stack to find the implicit
            namespace (if used, see `run_namespace`).
        namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        """
        from brian2.core.namespace import get_local_namespace  # avoids circular import

        if isinstance(item, str):
            # Check whether the group still exists to give a more meaningful
            # error message if it does not
            try:
                self.group.name
            except ReferenceError:
                raise ReferenceError(
                    "Cannot use string expressions, the "
                    f"group '{self.group_name}', providing the "
                    "context for the expression, no longer exists. "
                    "Consider holding an explicit reference "
                    "to it to keep it alive."
                )
            if namespace is None:
                namespace = get_local_namespace(level=level + 1)
            values = self.get_with_expression(item, run_namespace=namespace)
        else:
            if isinstance(self.variable, Subexpression):
                if namespace is None:
                    namespace = get_local_namespace(level=level + 1)
                values = self.get_subexpression_with_index_array(
                    item, run_namespace=namespace
                )
            else:
                values = self.get_with_index_array(item)

        if self.dim is DIMENSIONLESS or self.dim is None:
            return values
        else:
            return Quantity(values, self.dim)

    def __getitem__(self, item):
        return self.get_item(item, level=1)

    def set_item(self, item, value, level=0, namespace=None):
        """
        Set this variable. This function is called by `__setitem__` but there
        is also a situation where it should be called directly: if the context
        for string-based expressions is higher up in the stack, this function
        allows to set the `level` argument accordingly.

        Parameters
        ----------
        item : slice, `ndarray` or string
            The index for the setting operation
        value : `Quantity`, `ndarray` or number
            The value for the setting operation
        level : int, optional
            How much farther to go up in the stack to find the implicit
            namespace (if used, see `run_namespace`).
        namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        """
        from brian2.core.namespace import get_local_namespace  # avoids circular import

        variable = self.variable
        if variable.read_only:
            raise TypeError(f"Variable {self.name} is read-only.")

        # Check whether the group allows writing to the variable (e.g. for
        # synaptic variables, writing is only allowed after a connect)
        try:
            self.group.check_variable_write(variable)
        except ReferenceError:
            # Ignore problems with weakly referenced groups that don't exist
            # anymore at this time (e.g. when doing neuron.axon.var = ...)
            pass

        # The second part is equivalent to item == slice(None) but formulating
        # it this way prevents a FutureWarning if one of the elements is a
        # numpy array
        if isinstance(item, slice) and (
            item.start is None and item.stop is None and item.step is None
        ):
            item = "True"

        check_units = self.dim is not None

        if namespace is None:
            namespace = get_local_namespace(level=level + 1)

        # Both index and values are strings, use a single code object do deal
        # with this situation
        if isinstance(value, str) and isinstance(item, str):
            self.set_with_expression_conditional(
                item, value, check_units=check_units, run_namespace=namespace
            )
        elif isinstance(item, str):
            try:
                if isinstance(value, str):
                    raise TypeError  # Will be dealt with below
                value = np.asanyarray(value).item()
            except (TypeError, ValueError):
                if item != "True":
                    raise TypeError(
                        "When setting a variable based on a string "
                        "index, the value has to be a string or a "
                        "scalar."
                    )

            if item == "True":
                # We do not want to go through code generation for runtime
                self.set_with_index_array(slice(None), value, check_units=check_units)
            else:
                self.set_with_expression_conditional(
                    item, repr(value), check_units=check_units, run_namespace=namespace
                )
        elif isinstance(value, str):
            self.set_with_expression(
                item, value, check_units=check_units, run_namespace=namespace
            )
        else:  # No string expressions involved
            self.set_with_index_array(item, value, check_units=check_units)

    def __setitem__(self, item, value):
        self.set_item(item, value, level=1)

    @device_override("variableview_set_with_expression")
    def set_with_expression(self, item, code, run_namespace, check_units=True):
        """
        Sets a variable using a string expression. Is called by
        `VariableView.set_item` for statements such as
        ``S.var[:, :] = 'exp(-abs(i-j)/space_constant)*nS'``

        Parameters
        ----------
        item : `ndarray`
            The indices for the variable (in the context of this `group`).
        code : str
            The code that should be executed to set the variable values.
            Can contain references to indices, such as `i` or `j`
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        check_units : bool, optional
            Whether to check the units of the expression.
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        """
        # Some fairly complicated code to raise a warning in ambiguous
        # situations, when indexing with a group. For example, in:
        #   group.v[subgroup] =  'i'
        # the index 'i' is the index of 'group' ("absolute index") and not of
        # subgroup ("relative index")
        if hasattr(item, "variables") or (
            isinstance(item, tuple)
            and any(hasattr(one_item, "variables") for one_item in item)
        ):
            # Determine the variables that are used in the expression
            from brian2.codegen.translation import get_identifiers_recursively

            identifiers = get_identifiers_recursively([code], self.group.variables)
            variables = self.group.resolve_all(
                identifiers, run_namespace, user_identifiers=set()
            )
            if not isinstance(item, tuple):
                index_groups = [item]
            else:
                index_groups = item

            for varname, var in variables.items():
                for index_group in index_groups:
                    if not hasattr(index_group, "variables"):
                        continue
                    if (
                        varname in index_group.variables
                        or var.name in index_group.variables
                    ):
                        indexed_var = index_group.variables.get(
                            varname, index_group.variables.get(var.name)
                        )
                        if indexed_var is not var:
                            logger.warn(
                                "The string expression used for setting "
                                f"'{self.name}' refers to '{varname}' which "
                                "might be ambiguous. It will be "
                                "interpreted as referring to "
                                f"'{varname}' in '{self.group.name}', not as "
                                "a variable of a group used for "
                                "indexing.",
                                "ambiguous_string_expression",
                            )
                            break  # no need to warn more than once for a variable

        indices = np.atleast_1d(self.indexing(item))
        abstract_code = f"{self.name} = {code}"
        variables = Variables(self.group)
        variables.add_array(
            "_group_idx", size=len(indices), dtype=np.int32, values=indices
        )

        # TODO: Have an additional argument to avoid going through the index
        # array for situations where iterate_all could be used
        from brian2.codegen.codeobject import create_runner_codeobj
        from brian2.devices.device import get_device

        device = get_device()

        codeobj = create_runner_codeobj(
            self.group,
            abstract_code,
            "group_variable_set",
            additional_variables=variables,
            check_units=check_units,
            run_namespace=run_namespace,
            codeobj_class=device.code_object_class(
                fallback_pref="codegen.string_expression_target"
            ),
        )
        codeobj()

    @device_override("variableview_set_with_expression_conditional")
    def set_with_expression_conditional(
        self, cond, code, run_namespace, check_units=True
    ):
        """
        Sets a variable using a string expression and string condition. Is
        called by `VariableView.set_item` for statements such as
        ``S.var['i!=j'] = 'exp(-abs(i-j)/space_constant)*nS'``

        Parameters
        ----------
        cond : str
            The string condition for which the variables should be set.
        code : str
            The code that should be executed to set the variable values.
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        check_units : bool, optional
            Whether to check the units of the expression.
        """
        variable = self.variable
        if variable.scalar and cond != "True":
            raise IndexError(
                f"Cannot conditionally set the scalar variable '{self.name}'."
            )
        abstract_code_cond = f"_cond = {cond}"
        abstract_code = f"{self.name} = {code}"
        variables = Variables(None)
        variables.add_auxiliary_variable("_cond", dtype=bool)
        from brian2.codegen.codeobject import create_runner_codeobj

        # TODO: Have an additional argument to avoid going through the index
        # array for situations where iterate_all could be used
        from brian2.devices.device import get_device

        device = get_device()
        codeobj = create_runner_codeobj(
            self.group,
            {"condition": abstract_code_cond, "statement": abstract_code},
            "group_variable_set_conditional",
            additional_variables=variables,
            check_units=check_units,
            run_namespace=run_namespace,
            codeobj_class=device.code_object_class(
                fallback_pref="codegen.string_expression_target"
            ),
        )
        codeobj()

    @device_override("variableview_get_with_expression")
    def get_with_expression(self, code, run_namespace):
        """
        Gets a variable using a string expression. Is called by
        `VariableView.get_item` for statements such as
        ``print(G.v['g_syn > 0'])``.

        Parameters
        ----------
        code : str
            An expression that states a condition for elements that should be
            selected. Can contain references to indices, such as ``i`` or ``j``
            and to state variables. For example: ``'i>3 and v>0*mV'``.
        run_namespace : dict-like
            An additional namespace that is used for variable lookup (either
            an explicitly defined namespace or one taken from the local
            context).
        """
        variable = self.variable
        if variable.scalar:
            raise IndexError(
                f"Cannot access the variable '{self.name}' with a "
                "string expression, it is a scalar variable."
            )
        # Add the recorded variable under a known name to the variables
        # dictionary. Important to deal correctly with
        # the type of the variable in C++
        variables = Variables(None)
        variables.add_auxiliary_variable(
            "_variable",
            dimensions=variable.dim,
            dtype=variable.dtype,
            scalar=variable.scalar,
        )
        variables.add_auxiliary_variable("_cond", dtype=bool)

        abstract_code = f"_variable = {self.name}\n"
        abstract_code += f"_cond = {code}"
        from brian2.codegen.codeobject import create_runner_codeobj
        from brian2.devices.device import get_device

        device = get_device()
        codeobj = create_runner_codeobj(
            self.group,
            abstract_code,
            "group_variable_get_conditional",
            additional_variables=variables,
            run_namespace=run_namespace,
            codeobj_class=device.code_object_class(
                fallback_pref="codegen.string_expression_target"
            ),
        )
        return codeobj()

    @device_override("variableview_get_with_index_array")
    def get_with_index_array(self, item):
        variable = self.variable
        if variable.scalar:
            if not (isinstance(item, slice) and item == slice(None)):
                raise IndexError(
                    f"Illegal index for variable '{self.name}', it is a "
                    "scalar variable."
                )
            indices = 0
        elif (
            isinstance(item, slice) and item == slice(None) and self.index_var == "_idx"
        ):
            indices = slice(None)
        # Quick fix for matplotlib calling 1-d variables as var[:, np.newaxis]
        # The test is a bit verbose, but we need to avoid comparisons that raise errors
        # (e.g. comparing an array to slice(None))
        elif (
            isinstance(item, tuple)
            and len(item) == 2
            and isinstance(item[0], slice)
            and item[0] == slice(None)
            and item[1] is None
        ):
            if self.index_var == "_idx":
                return variable.get_value()[item]
            else:
                return variable.get_value()[self.index_var.get_value()][item]
        else:
            indices = self.indexing(item, self.index_var)

        return variable.get_value()[indices]

    @device_override("variableview_get_subexpression_with_index_array")
    def get_subexpression_with_index_array(self, item, run_namespace):
        variable = self.variable
        if variable.scalar:
            if not (isinstance(item, slice) and item == slice(None)):
                raise IndexError(
                    f"Illegal index for variable '{self.name}', it is a "
                    "scalar variable."
                )
            indices = np.array(0)
        else:
            indices = self.indexing(item, self.index_var)

        # For "normal" variables, we can directly access the underlying data
        # and use the usual slicing syntax. For subexpressions, however, we
        # have to evaluate code for the given indices
        variables = Variables(None, default_index="_group_index")
        variables.add_auxiliary_variable(
            "_variable",
            dimensions=variable.dim,
            dtype=variable.dtype,
            scalar=variable.scalar,
        )
        if indices.shape == ():
            single_index = True
            indices = np.array([indices])
        else:
            single_index = False
        variables.add_array("_group_idx", size=len(indices), dtype=np.int32)
        variables["_group_idx"].set_value(indices)
        # Force the use of this variable as a replacement for the original
        # index variable
        using_orig_index = [
            varname
            for varname, index in self.group.variables.indices.items()
            if index == self.index_var_name and index != "0"
        ]
        for varname in using_orig_index:
            variables.indices[varname] = "_idx"

        abstract_code = f"_variable = {self.name}\n"
        from brian2.codegen.codeobject import create_runner_codeobj
        from brian2.devices.device import get_device

        device = get_device()
        codeobj = create_runner_codeobj(
            self.group,
            abstract_code,
            "group_variable_get",
            # Setting the user code to an empty
            # string suppresses warnings if the
            # subexpression refers to variable
            # names that are also present in the
            # local namespace
            user_code="",
            needed_variables=["_group_idx"],
            additional_variables=variables,
            run_namespace=run_namespace,
            codeobj_class=device.code_object_class(
                fallback_pref="codegen.string_expression_target"
            ),
        )
        result = codeobj()
        if single_index and not variable.scalar:
            return result[0]
        else:
            return result

    @device_override("variableview_set_with_index_array")
    def set_with_index_array(self, item, value, check_units):
        variable = self.variable
        if check_units:
            fail_for_dimension_mismatch(
                variable.dim, value, f"Incorrect unit for setting variable {self.name}"
            )
        if variable.scalar:
            if not (isinstance(item, slice) and item == slice(None)):
                raise IndexError(
                    "Illegal index for variable '{self.name}', it is a scalar variable."
                )
            indices = 0
        elif (
            isinstance(item, slice) and item == slice(None) and self.index_var == "_idx"
        ):
            indices = slice(None)
        else:
            indices = self.indexing(item, self.index_var)

            q = Quantity(value)
            if len(q.shape):
                if not len(q.shape) == 1 or len(q) != 1 and len(q) != len(indices):
                    raise ValueError(
                        "Provided values do not match the size "
                        "of the indices, "
                        f"{len(q)} != {len(indices)}."
                    )
        variable.get_value()[indices] = value

    # Allow some basic calculations directly on the ArrayView object
    def __array__(self, dtype=None, copy=None):
        try:
            # This will fail for subexpressions that refer to external
            # parameters
            values = self[:]
            # Never hand over copy = None
            return np.array(values, dtype=dtype, copy=copy is not False, subok=True)
        except ValueError:
            var = self.variable.name
            raise ValueError(
                f"Cannot get the values for variable {var}. If it "
                "is a subexpression referring to external "
                f"variables, use 'group.{var}[:]' instead of "
                f"'group.{var}'"
            )

    def __array__ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method == "__call__":
            return ufunc(self[:], *inputs, **kwargs)
        else:
            return NotImplemented

    def __len__(self):
        return len(self.get_item(slice(None), level=1))

    def __neg__(self):
        return -self.get_item(slice(None), level=1)

    def __pos__(self):
        return self.get_item(slice(None), level=1)

    def __add__(self, other):
        return self.get_item(slice(None), level=1) + np.asanyarray(other)

    def __radd__(self, other):
        return np.asanyarray(other) + self.get_item(slice(None), level=1)

    def __sub__(self, other):
        return self.get_item(slice(None), level=1) - np.asanyarray(other)

    def __rsub__(self, other):
        return np.asanyarray(other) - self.get_item(slice(None), level=1)

    def __mul__(self, other):
        return self.get_item(slice(None), level=1) * np.asanyarray(other)

    def __rmul__(self, other):
        return np.asanyarray(other) * self.get_item(slice(None), level=1)

    def __div__(self, other):
        return self.get_item(slice(None), level=1) / np.asanyarray(other)

    def __truediv__(self, other):
        return self.get_item(slice(None), level=1) / np.asanyarray(other)

    def __floordiv__(self, other):
        return self.get_item(slice(None), level=1) // np.asanyarray(other)

    def __rdiv__(self, other):
        return np.asanyarray(other) / self.get_item(slice(None), level=1)

    def __rtruediv__(self, other):
        return np.asanyarray(other) / self.get_item(slice(None), level=1)

    def __rfloordiv__(self, other):
        return np.asanyarray(other) // self.get_item(slice(None), level=1)

    def __mod__(self, other):
        return self.get_item(slice(None), level=1) % np.asanyarray(other)

    def __pow__(self, power, modulo=None):
        if modulo is not None:
            return self.get_item(slice(None), level=1) ** power % modulo
        else:
            return self.get_item(slice(None), level=1) ** power

    def __rpow__(self, other):
        if self.dim is not DIMENSIONLESS:
            raise TypeError(
                f"Cannot use '{self.name}' as an exponent, it has "
                f"dimensions {get_unit_for_display(self.unit)}."
            )
        return other ** self.get_item(slice(None), level=1)

    def __iadd__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var + expression' "
                "instead of group.var += 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] + np.asanyarray(other)
        self[:] = rhs
        return self

    # Support matrix multiplication with @
    def __matmul__(self, other):
        return self.get_item(slice(None), level=1) @ np.asanyarray(other)

    def __rmatmul__(self, other):
        return np.asanyarray(other) @ self.get_item(slice(None), level=1)

    def __isub__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var - expression' "
                "instead of group.var -= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] - np.asanyarray(other)
        self[:] = rhs
        return self

    def __imul__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var * expression' "
                "instead of group.var *= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] * np.asanyarray(other)
        self[:] = rhs
        return self

    def __idiv__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var / expression' "
                "instead of group.var /= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] / np.asanyarray(other)
        self[:] = rhs
        return self

    def __ifloordiv__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var // expression' "
                "instead of group.var //= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] // np.asanyarray(other)
        self[:] = rhs
        return self

    def __imod__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var // expression' "
                "instead of group.var //= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] % np.asanyarray(other)
        self[:] = rhs
        return self

    def __ipow__(self, other):
        if isinstance(other, str):
            raise TypeError(
                "In-place modification with strings not "
                "supported. Use group.var = 'var ** expression' "
                "instead of group.var **= 'expression'."
            )
        elif isinstance(self.variable, Subexpression):
            raise TypeError("Cannot assign to a subexpression.")
        else:
            rhs = self[:] ** np.asanyarray(other)
        self[:] = rhs
        return self

    # Also allow logical comparisons

    def __eq__(self, other):
        return self.get_item(slice(None), level=1) == np.asanyarray(other)

    def __ne__(self, other):
        return self.get_item(slice(None), level=1) != np.asanyarray(other)

    def __lt__(self, other):
        return self.get_item(slice(None), level=1) < np.asanyarray(other)

    def __le__(self, other):
        return self.get_item(slice(None), level=1) <= np.asanyarray(other)

    def __gt__(self, other):
        return self.get_item(slice(None), level=1) > np.asanyarray(other)

    def __ge__(self, other):
        return self.get_item(slice(None), level=1) >= np.asanyarray(other)

    def __repr__(self):
        varname = self.name
        if self.dim is None:
            varname += "_"

        if self.variable.scalar:
            dim = self.dim if self.dim is not None else DIMENSIONLESS
            values = repr(Quantity(self.variable.get_value().item(), dim=dim))
        else:
            try:
                # This will fail for subexpressions that refer to external
                # parameters
                values = repr(self[:])
            except KeyError:
                values = (
                    "[Subexpression refers to external parameters. Use "
                    f"'group.{self.variable.name}[:]']"
                )

        return f"<{self.group_name}.{varname}: {values}>"

    def __hash__(self):
        return hash((self.group_name, self.name))

    # Get access to some basic properties of the underlying array
    @property
    def shape(self):
        if self.ndim == 1:
            if not self.variable.scalar:
                # This is safer than using the variable size, since it also works for subgroups
                # see GitHub issue #1555
                size = self.group.stop - self.group.start
                assert size <= self.variable.size
            else:
                size = self.variable.size

            return (size,)
        else:
            return self.variable.size

    @property
    def ndim(self):
        return getattr(self.variable, "ndim", 1)

    @property
    def dtype(self):
        return self.variable.dtype


class Variables(Mapping):
    """
    A container class for storing `Variable` objects. Instances of this class
    are used as the `Group.variables` attribute and can be accessed as
    (read-only) dictionaries.

    Parameters
    ----------
    owner : `Nameable`
        The object (typically a `Group`) "owning" the variables.
    default_index : str, optional
        The index to use for the variables (only relevant for `ArrayVariable`
        and `DynamicArrayVariable`). Defaults to ``'_idx'``.
    """

    def __init__(self, owner, default_index="_idx"):
        #: A reference to the `Group` owning these variables
        self.owner = weakproxy_with_fallback(owner)
        # The index that is used for arrays if no index is given explicitly
        self.default_index = default_index

        # We do the import here to avoid a circular dependency.
        from brian2.devices.device import get_device

        self.device = get_device()

        self._variables = {}
        #: A dictionary given the index name for every array name
        self.indices = collections.defaultdict(functools.partial(str, default_index))
        # Note that by using functools.partial (instead of e.g. a lambda
        # function) above, this object remains pickable.

    def __getstate__(self):
        state = self.__dict__.copy()
        state["owner"] = state["owner"].__repr__.__self__
        return state

    def __setstate__(self, state):
        state["owner"] = weakproxy_with_fallback(state["owner"])
        self.__dict__ = state

    def __getitem__(self, item):
        return self._variables[item]

    def __len__(self):
        return len(self._variables)

    def __iter__(self):
        return iter(self._variables)

    def _add_variable(self, name, var, index=None):
        if name in self._variables:
            raise KeyError(
                f"The name '{name}' is already present in the variables dictionary."
            )
        # TODO: do some check for the name, part of it has to be device-specific
        self._variables[name] = var

        if isinstance(var, ArrayVariable):
            # Tell the device to actually create the array (or note it down for
            # later code generation in standalone).
            self.device.add_array(var)

        if getattr(var, "scalar", False):
            if index not in (None, "0"):
                raise ValueError("Cannot set an index for a scalar variable")
            self.indices[name] = "0"

        if index is not None:
            self.indices[name] = index

    def add_array(
        self,
        name,
        size,
        dimensions=DIMENSIONLESS,
        values=None,
        dtype=None,
        constant=False,
        read_only=False,
        scalar=False,
        unique=False,
        index=None,
    ):
        """
        Add an array (initialized with zeros).

        Parameters
        ----------
        name : str
            The name of the variable.
        dimensions : `Dimension`, optional
            The physical dimensions of the variable.
        size : int
            The size of the array.
        values : `ndarray`, optional
            The values to initalize the array with. If not specified, the array
            is initialized to zero.
        dtype : `dtype`, optional
            The dtype used for storing the variable. If none is given, defaults
            to `core.default_float_dtype`.
        constant : bool, optional
            Whether the variable's value is constant during a run.
            Defaults to ``False``.
        scalar : bool, optional
            Whether this is a scalar variable. Defaults to ``False``, if set to
            ``True``, also implies that `size` equals 1.
        read_only : bool, optional
            Whether this is a read-only variable, i.e. a variable that is set
            internally and cannot be changed by the user. Defaults
            to ``False``.
        index : str, optional
            The index to use for this variable. Defaults to
            `Variables.default_index`.
        unique : bool, optional
            See `ArrayVariable`. Defaults to ``False``.
        """
        if np.asanyarray(size).shape == ():
            # We want a basic Python type for the size instead of something
            # like numpy.int64
            size = int(size)
        var = ArrayVariable(
            name=name,
            dimensions=dimensions,
            owner=self.owner,
            device=self.device,
            size=size,
            dtype=dtype,
            constant=constant,
            scalar=scalar,
            read_only=read_only,
            unique=unique,
        )
        self._add_variable(name, var, index)
        # This could be avoided, but we currently need it so that standalone
        # allocates the memory
        self.device.init_with_zeros(var, dtype)
        if values is not None:
            if scalar:
                if np.asanyarray(values).shape != ():
                    raise ValueError("Need a scalar value.")
                self.device.fill_with_array(var, values)
            else:
                if len(values) != size:
                    raise ValueError(
                        "Size of the provided values does not match "
                        f"size: {len(values)} != {size}"
                    )
                self.device.fill_with_array(var, values)

    def add_arrays(
        self,
        names,
        size,
        dimensions=DIMENSIONLESS,
        dtype=None,
        constant=False,
        read_only=False,
        scalar=False,
        unique=False,
        index=None,
    ):
        """
        Adds several arrays (initialized with zeros) with the same attributes
        (size, units, etc.).

        Parameters
        ----------
        names : list of str
            The names of the variable.
        dimensions : `Dimension`, optional
            The physical dimensions of the variable.
        size : int
            The sizes of the arrays.
        dtype : `dtype`, optional
            The dtype used for storing the variables. If none is given, defaults
            to `core.default_float_dtype`.
        constant : bool, optional
            Whether the variables' values are constant during a run.
            Defaults to ``False``.
        scalar : bool, optional
            Whether these are scalar variables. Defaults to ``False``, if set to
            ``True``, also implies that `size` equals 1.
        read_only : bool, optional
            Whether these are read-only variables, i.e. variables that are set
            internally and cannot be changed by the user. Defaults
            to ``False``.
        index : str, optional
            The index to use for these variables. Defaults to
            `Variables.default_index`.
        unique : bool, optional
            See `ArrayVariable`. Defaults to ``False``.
        """
        for name in names:
            self.add_array(
                name,
                dimensions=dimensions,
                size=size,
                dtype=dtype,
                constant=constant,
                read_only=read_only,
                scalar=scalar,
                unique=unique,
                index=index,
            )

    def add_dynamic_array(
        self,
        name,
        size,
        dimensions=DIMENSIONLESS,
        values=None,
        dtype=None,
        constant=False,
        needs_reference_update=False,
        resize_along_first=False,
        read_only=False,
        unique=False,
        scalar=False,
        index=None,
    ):
        """
        Add a dynamic array.

        Parameters
        ----------
        name : str
            The name of the variable.
        dimensions : `Dimension`, optional
            The physical dimensions of the variable.
        size : int or tuple of int
            The (initital) size of the array.
        values : `ndarray`, optional
            The values to initalize the array with. If not specified, the array
            is initialized to zero.
        dtype : `dtype`, optional
            The dtype used for storing the variable. If none is given, defaults
            to `core.default_float_dtype`.
        constant : bool, optional
            Whether the variable's value is constant during a run.
            Defaults to ``False``.
        needs_reference_update : bool, optional
            Whether the code objects need a new reference to the underlying data at
            every time step. This should be set if the size of the array can be
            changed by other code objects. Defaults to ``False``.
        scalar : bool, optional
            Whether this is a scalar variable. Defaults to ``False``, if set to
            ``True``, also implies that `size` equals 1.
        read_only : bool, optional
            Whether this is a read-only variable, i.e. a variable that is set
            internally and cannot be changed by the user. Defaults
            to ``False``.
        index : str, optional
            The index to use for this variable. Defaults to
            `Variables.default_index`.
        unique : bool, optional
            See `DynamicArrayVariable`. Defaults to ``False``.
        """
        var = DynamicArrayVariable(
            name=name,
            dimensions=dimensions,
            owner=self.owner,
            device=self.device,
            size=size,
            dtype=dtype,
            constant=constant,
            needs_reference_update=needs_reference_update,
            resize_along_first=resize_along_first,
            scalar=scalar,
            read_only=read_only,
            unique=unique,
        )
        self._add_variable(name, var, index)
        if np.prod(size) > 0:
            self.device.resize(var, size)
        if values is None and np.prod(size) > 0:
            self.device.init_with_zeros(var, dtype)
        elif values is not None:
            if len(values) != size:
                raise ValueError(
                    "Size of the provided values does not match "
                    f"size: {len(values)} != {size}"
                )
            if np.prod(size) > 0:
                self.device.fill_with_array(var, values)

    def add_arange(
        self,
        name,
        size,
        start=0,
        dtype=np.int32,
        constant=True,
        read_only=True,
        unique=True,
        index=None,
    ):
        """
        Add an array, initialized with a range of integers.

        Parameters
        ----------
        name : str
            The name of the variable.
        size : int
            The size of the array.
        start : int
            The start value of the range.
        dtype : `dtype`, optional
            The dtype used for storing the variable. If none is given, defaults
            to `np.int32`.
        constant : bool, optional
            Whether the variable's value is constant during a run.
            Defaults to ``True``.
        read_only : bool, optional
            Whether this is a read-only variable, i.e. a variable that is set
            internally and cannot be changed by the user. Defaults
            to ``True``.
        index : str, optional
            The index to use for this variable. Defaults to
            `Variables.default_index`.
        unique : bool, optional
            See `ArrayVariable`. Defaults to ``True`` here.
        """
        self.add_array(
            name=name,
            dimensions=DIMENSIONLESS,
            size=size,
            dtype=dtype,
            constant=constant,
            read_only=read_only,
            unique=unique,
            index=index,
        )
        self.device.init_with_arange(self._variables[name], start, dtype=dtype)

    def add_constant(self, name, value, dimensions=DIMENSIONLESS):
        """
        Add a scalar constant (e.g. the number of neurons `N`).

        Parameters
        ----------
        name : str
            The name of the variable
        value: reference to the variable value
            The value of the constant.
        dimensions : `Dimension`, optional
            The physical dimensions of the variable. Note that the variable
            itself (as referenced by value) should never have units attached.
        """
        var = Constant(name=name, dimensions=dimensions, owner=self.owner, value=value)
        self._add_variable(name, var)

    def add_subexpression(
        self, name, expr, dimensions=DIMENSIONLESS, dtype=None, scalar=False, index=None
    ):
        """
        Add a named subexpression.

        Parameters
        ----------
        name : str
            The name of the subexpression.
        dimensions : `Dimension`
            The physical dimensions of the subexpression.
        expr : str
            The subexpression itself.
        dtype : `dtype`, optional
            The dtype used for the expression. Defaults to
            `core.default_float_dtype`.
        scalar : bool, optional
            Whether this is an expression only referring to scalar variables.
            Defaults to ``False``
        index : str, optional
            The index to use for this variable. Defaults to
            `Variables.default_index`.
        """
        var = Subexpression(
            name=name,
            dimensions=dimensions,
            expr=expr,
            owner=self.owner,
            dtype=dtype,
            device=self.device,
            scalar=scalar,
        )
        self._add_variable(name, var, index=index)

    def add_auxiliary_variable(
        self, name, dimensions=DIMENSIONLESS, dtype=None, scalar=False
    ):
        """
        Add an auxiliary variable (most likely one that is added automatically
        to abstract code, e.g. ``_cond`` for a threshold condition),
        specifying its type and unit for code generation.

        Parameters
        ----------
        name : str
            The name of the variable
        dimensions : `Dimension`
            The physical dimensions of the variable.
        dtype : `dtype`, optional
            The dtype used for storing the variable. If none is given, defaults
            to `core.default_float_dtype`.
        scalar : bool, optional
            Whether the variable is a scalar value (``True``) or vector-valued,
            e.g. defined for every neuron (``False``). Defaults to ``False``.
        """
        var = AuxiliaryVariable(
            name=name, dimensions=dimensions, dtype=dtype, scalar=scalar
        )
        self._add_variable(name, var)

    def add_referred_subexpression(self, name, group, subexpr, index):
        identifiers = subexpr.identifiers
        substitutions = {}
        for identifier in identifiers:
            if identifier not in subexpr.owner.variables:
                # external variable --> nothing to do
                continue
            subexpr_var = subexpr.owner.variables[identifier]
            if hasattr(subexpr_var, "owner"):
                new_name = f"_{name}_{subexpr.owner.name}_{identifier}"
            else:
                new_name = f"_{name}_{identifier}"
            substitutions[identifier] = new_name

            subexpr_var_index = group.variables.indices[identifier]
            if subexpr_var_index == group.variables.default_index:
                subexpr_var_index = index
            elif subexpr_var_index == "0":
                pass  # nothing to do for a shared variable
            elif subexpr_var_index == index:
                pass  # The same index as the main subexpression
            elif index != self.default_index:
                index_var = self._variables.get(index, None)
                if isinstance(index_var, DynamicArrayVariable):
                    raise TypeError(
                        f"Cannot link to subexpression '{name}': it refers "
                        f"to the variable '{identifier}' which is indexed "
                        f"with the dynamic index '{subexpr_var_index}'."
                    )
            else:
                self.add_reference(subexpr_var_index, group)

            self.indices[new_name] = subexpr_var_index

            if isinstance(subexpr_var, Subexpression):
                self.add_referred_subexpression(
                    new_name, group, subexpr_var, subexpr_var_index
                )
            else:
                self.add_reference(new_name, group, identifier, subexpr_var_index)

        new_expr = word_substitute(subexpr.expr, substitutions)
        new_subexpr = Subexpression(
            name,
            self.owner,
            new_expr,
            dimensions=subexpr.dim,
            device=subexpr.device,
            dtype=subexpr.dtype,
            scalar=subexpr.scalar,
        )
        self._variables[name] = new_subexpr

    def add_reference(self, name, group, varname=None, index=None):
        """
        Add a reference to a variable defined somewhere else (possibly under
        a different name). This is for example used in `Subgroup` and
        `Synapses` to refer to variables in the respective `NeuronGroup`.

        Parameters
        ----------
        name : str
            The name of the variable (in this group, possibly a different name
            from `var.name`).
        group : `Group`
            The group from which `var` is referenced
        varname : str, optional
            The variable to refer to. If not given, defaults to `name`.
        index : str, optional
            The index that should be used for this variable (defaults to
            `Variables.default_index`).
        """
        if varname is None:
            varname = name
        if varname not in group.variables:
            raise KeyError(f"Group {group.name} does not have a variable {varname}.")
        if index is None:
            if group.variables[varname].scalar:
                index = "0"
            else:
                index = self.default_index

        if (
            self.owner is not None
            and self.owner.name != group.name
            and index in self.owner.variables
        ):
            if (
                not self.owner.variables[index].read_only
                or isinstance(self.owner.variables[index], DynamicArrayVariable)
            ) and group.variables.indices[varname] != group.variables.default_index:
                raise TypeError(
                    f"Cannot link variable '{name}' to '{varname}' in "
                    f"group '{group.name}' -- need to precalculate "
                    f"direct indices but index {index} can change"
                )

        # We don't overwrite existing names with references
        if name not in self._variables:
            var = group.variables[varname]
            if isinstance(var, Subexpression):
                self.add_referred_subexpression(name, group, var, index)
            else:
                self._variables[name] = var
            self.indices[name] = index

    def add_references(self, group, varnames, index=None):
        """
        Add all `Variable` objects from a name to `Variable` mapping with the
        same name as in the original mapping.

        Parameters
        ----------
        group : `Group`
            The group from which the `variables` are referenced
        varnames : iterable of str
            The variables that should be referred to in the current group
        index : str, optional
            The index to use for all the variables (defaults to
            `Variables.default_index`)
        """
        for name in varnames:
            self.add_reference(name, group, name, index)

    def add_object(self, name, obj):
        """
        Add an arbitrary Python object. This is only meant for internal use
        and therefore only names starting with an underscore are allowed.

        Parameters
        ----------
        name : str
            The name used for this object (has to start with an underscore).
        obj : object
            An arbitrary Python object that needs to be accessed directly from
            a `CodeObject`.
        """
        if not name.startswith("_"):
            raise ValueError(
                "This method is only meant for internally used "
                "objects, the name therefore has to start with "
                "an underscore"
            )
        self._variables[name] = obj

    def create_clock_variables(self, clock, prefix=""):
        """
        Convenience function to add the ``t`` and ``dt`` attributes of a
        `clock`.

        Parameters
        ----------
        clock : `Clock`
            The clock that should be used for ``t`` and ``dt``.
        prefix : str, optional
            A prefix for the variable names. Used for example in monitors to
            not confuse the dynamic array of recorded times with the current
            time in the recorded group.
        """
        self.add_reference(f"{prefix}t", clock, "t")
        self.add_reference(f"{prefix}dt", clock, "dt")
        self.add_reference(f"{prefix}t_in_timesteps", clock, "timestep")



"""
TODO: rewrite this (verbatim from Brian 1.x), more efficiency
"""

import numpy as np

__all__ = ["DynamicArray", "DynamicArray1D"]


def getslices(shape, from_start=True):
    if from_start:
        return tuple(slice(0, x) for x in shape)
    else:
        return tuple(slice(x, None) for x in shape)


class DynamicArray:
    """
    An N-dimensional dynamic array class

    The array can be resized in any dimension, and the class will handle
    allocating a new block of data and copying when necessary.

    .. warning::
        The data will NOT be contiguous for >1D arrays. To ensure this, you will
        either need to use 1D arrays, or to copy the data, or use the shrink
        method with the current size (although note that in both cases you
        negate the memory and efficiency benefits of the dynamic array).

    Initialisation arguments:

    ``shape``, ``dtype``
        The shape and dtype of the array to initialise, as in Numpy. For 1D
        arrays, shape can be a single int, for ND arrays it should be a tuple.
    ``factor``
        The resizing factor (see notes below). Larger values tend to lead to
        more wasted memory, but more computationally efficient code.
    ``use_numpy_resize``, ``refcheck``
        Normally, when you resize the array it creates a new array and copies
        the data. Sometimes, it is possible to resize an array without a copy,
        and if this option is set it will attempt to do this. However, this can
        cause memory problems if you are not careful so the option is off by
        default. You need to ensure that you do not create slices of the array
        so that no references to the memory exist other than the main array
        object. If you are sure you know what you're doing, you can switch this
        reference check off. Note that resizing in this way is only done if you
        resize in the first dimension.

    The array is initialised with zeros. The data is stored in the attribute
    ``data`` which is a Numpy array.


    Some numpy methods are implemented and can work directly on the array object,
    including ``len(arr)``, ``arr[...]`` and ``arr[...]=...``. In other cases,
    use the ``data`` attribute.

    Examples
    --------

    >>> x = DynamicArray((2, 3), dtype=int)
    >>> x[:] = 1
    >>> x.resize((3, 3))
    >>> x[:] += 1
    >>> x.resize((3, 4))
    >>> x[:] += 1
    >>> x.resize((4, 4))
    >>> x[:] += 1
    >>> x.data[:] = x.data**2
    >>> x.data
    array([[16, 16, 16,  4],
           [16, 16, 16,  4],
           [ 9,  9,  9,  4],
           [ 1,  1,  1,  1]])

    Notes
    -----

    The dynamic array returns a ``data`` attribute which is a view on the larger
    ``_data`` attribute. When a resize operation is performed, and a specific
    dimension is enlarged beyond the size in the ``_data`` attribute, the size
    is increased to the larger of ``cursize*factor`` and ``newsize``. This
    ensures that the amortized cost of increasing the size of the array is O(1).
    """

    def __init__(
        self, shape, dtype=float, factor=2, use_numpy_resize=False, refcheck=True
    ):
        if isinstance(shape, int):
            shape = (shape,)
        self._data = np.zeros(shape, dtype=dtype)
        self.data = self._data
        self.dtype = dtype
        self.shape = self._data.shape
        self.factor = factor
        self.use_numpy_resize = use_numpy_resize
        self.refcheck = refcheck

    def resize(self, newshape):
        """
        Resizes the data to the new shape, which can be a different size to the
        current data, but should have the same rank, i.e. same number of
        dimensions.
        """
        datashapearr = np.array(self._data.shape)
        newshapearr = np.array(newshape)
        resizedimensions = newshapearr > datashapearr
        if resizedimensions.any():
            # resize of the data is needed
            minnewshapearr = datashapearr  # .copy()
            dimstoinc = minnewshapearr[resizedimensions]
            incdims = np.array(dimstoinc * self.factor, dtype=int)
            newdims = np.maximum(incdims, dimstoinc + 1)
            minnewshapearr[resizedimensions] = newdims
            newshapearr = np.maximum(newshapearr, minnewshapearr)
            do_resize = False
            if self.use_numpy_resize and self._data.flags["C_CONTIGUOUS"]:
                if sum(resizedimensions) == resizedimensions[0]:
                    do_resize = True
            if do_resize:
                self.data = None
                self._data.resize(tuple(newshapearr), refcheck=self.refcheck)
            else:
                newdata = np.zeros(tuple(newshapearr), dtype=self.dtype)
                slices = getslices(self._data.shape)
                newdata[slices] = self._data
                self._data = newdata
        elif (newshapearr < self.shape).any():
            # If we reduced the size, set the no longer used memory to 0
            self._data[getslices(newshape, from_start=False)] = 0
        # Reduce our view to the requested size if necessary
        self.data = self._data[getslices(newshape, from_start=True)]
        self.shape = self.data.shape

    def resize_along_first(self, newshape):
        new_dimension = newshape[0]
        if new_dimension > self._data.shape[0]:
            new_size = np.maximum(self._data.shape[0] * self.factor, new_dimension + 1)
            final_new_shape = np.array(self._data.shape)
            final_new_shape[0] = new_size
            if self.use_numpy_resize and self._data.flags["C_CONTIGUOUS"]:
                self.data = None
                self._data.resize(tuple(final_new_shape), refcheck=self.refcheck)
            else:
                newdata = np.zeros(tuple(final_new_shape), dtype=self.dtype)
                slices = getslices(self._data.shape)
                newdata[slices] = self._data
                self._data = newdata
        elif newshape < self.shape:
            # If we reduced the size, set the no longer used memory to 0
            self._data[new_dimension:] = 0
        # Reduce our view to the requested size if necessary
        self.data = self._data[:new_dimension]
        self.shape = newshape

    def shrink(self, newshape):
        """
        Reduces the data to the given shape, which should be smaller than the
        current shape. `resize` can also be used with smaller values, but
        it will not shrink the allocated memory, whereas `shrink` will
        reallocate the memory. This method should only be used infrequently, as
        if it is used frequently it will negate the computational efficiency
        benefits of the DynamicArray.
        """
        if isinstance(newshape, int):
            newshape = (newshape,)
        shapearr = np.array(self.shape)
        newshapearr = np.array(newshape)
        if (newshapearr <= shapearr).all():
            newdata = np.zeros(newshapearr, dtype=self.dtype)
            newdata[:] = self._data[getslices(newshapearr)]
            self._data = newdata
            self.shape = tuple(newshapearr)
            self.data = self._data

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def __setitem__(self, item, val):
        self.data.__setitem__(item, val)

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return self.data.__str__()

    def __repr__(self):
        return self.data.__repr__()


class DynamicArray1D(DynamicArray):
    """
    Version of `DynamicArray` with specialised ``resize`` method designed
    to be more efficient.
    """

    def resize(self, newshape):
        (datashape,) = self._data.shape
        if newshape > datashape:
            (shape,) = self.shape  # we work with int shapes only
            newdatashape = max(newshape, int(shape * self.factor) + 1)
            if self.use_numpy_resize and self._data.flags["C_CONTIGUOUS"]:
                self.data = None
                self._data.resize(newdatashape, refcheck=self.refcheck)
            else:
                newdata = np.zeros(newdatashape, dtype=self.dtype)
                newdata[:shape] = self.data
                self._data = newdata
        elif newshape < self.shape[0]:
            # If we reduced the size, set the no longer used memory to 0
            self._data[newshape:] = 0
        # Reduce our view to the requested size if necessary
        self.data = self._data[:newshape]
        self.shape = (newshape,)



import itertools

import numpy as np

from brian2.core.functions import DEFAULT_FUNCTIONS, timestep
from brian2.core.variables import ArrayVariable
from brian2.parsing.bast import brian_dtype_from_dtype
from brian2.parsing.rendering import NumpyNodeRenderer
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers, indent, word_substitute

from .base import CodeGenerator

__all__ = ["NumpyCodeGenerator"]


logger = get_logger(__name__)


class VectorisationError(Exception):
    pass


class NumpyCodeGenerator(CodeGenerator):
    """
    Numpy language

    Essentially Python but vectorised.
    """

    class_name = "numpy"

    _use_ufunc_at_vectorisation = True  # allow this to be off for testing only

    def translate_expression(self, expr):
        expr = word_substitute(expr, self.func_name_replacements)
        return (
            NumpyNodeRenderer(auto_vectorise=self.auto_vectorise)
            .render_expr(expr, self.variables)
            .strip()
        )

    def translate_statement(self, statement):
        # TODO: optimisation, translate arithmetic to a sequence of inplace
        # operations like a=b+c -> add(b, c, a)
        var, op, expr, comment = (
            statement.var,
            statement.op,
            statement.expr,
            statement.comment,
        )
        if op == ":=":
            op = "="
        # For numpy we replace complex expressions involving a single boolean variable into a
        # where(boolvar, expr_if_true, expr_if_false)
        if (
            statement.used_boolean_variables is not None
            and len(statement.used_boolean_variables) == 1
            and brian_dtype_from_dtype(statement.dtype) == "float"
            and statement.complexity_std > sum(statement.complexities.values())
        ):
            used_boolvars = statement.used_boolean_variables
            bool_simp = statement.boolean_simplified_expressions
            boolvar = used_boolvars[0]
            for bool_assigns, simp_expr in bool_simp.items():
                _, boolval = bool_assigns[0]
                if boolval:
                    expr_true = simp_expr
                else:
                    expr_false = simp_expr
            code = f"{var} {op} _numpy.where({boolvar}, {expr_true}, {expr_false})"
        else:
            code = f"{var} {op} {self.translate_expression(expr)}"
        if len(comment):
            code += f" # {comment}"
        return code

    def ufunc_at_vectorisation(
        self,
        statement,
        variables,
        indices,
        conditional_write_vars,
        created_vars,
        used_variables,
    ):
        if not self._use_ufunc_at_vectorisation:
            raise VectorisationError()
        # Avoids circular import
        from brian2.devices.device import device

        # See https://github.com/brian-team/brian2/pull/531 for explanation
        used = set(get_identifiers(statement.expr))
        used = used.intersection(
            k for k in list(variables.keys()) if k in indices and indices[k] != "_idx"
        )
        used_variables.update(used)
        if statement.var in used_variables:
            raise VectorisationError()
        expr = NumpyNodeRenderer(auto_vectorise=self.auto_vectorise).render_expr(
            statement.expr
        )

        if (
            statement.op == ":="
            or indices[statement.var] == "_idx"
            or not statement.inplace
        ):
            if statement.op == ":=":
                op = "="
            else:
                op = statement.op
            line = f"{statement.var} {op} {expr}"
        elif statement.inplace:
            if statement.op == "+=":
                ufunc_name = "_numpy.add"
            elif statement.op == "*=":
                ufunc_name = "_numpy.multiply"
            elif statement.op == "/=":
                ufunc_name = "_numpy.divide"
            elif statement.op == "-=":
                ufunc_name = "_numpy.subtract"
            else:
                raise VectorisationError()
            array_name = device.get_array_name(variables[statement.var])
            idx = indices[statement.var]
            line = f"{ufunc_name}.at({array_name}, {idx}, {expr})"
            line = self.conditional_write(
                line,
                statement,
                variables,
                conditional_write_vars=conditional_write_vars,
                created_vars=created_vars,
            )
        else:
            raise VectorisationError()

        if len(statement.comment):
            line += f" # {statement.comment}"

        return line

    def vectorise_code(self, statements, variables, variable_indices, index="_idx"):
        created_vars = {stmt.var for stmt in statements if stmt.op == ":="}
        try:
            lines = []
            used_variables = set()
            for statement in statements:
                lines.append(
                    "#  Abstract code: "
                    f" {statement.var} {statement.op} {statement.expr}"
                )
                # We treat every statement individually with its own read and write code
                # to be on the safe side
                read, write, indices, conditional_write_vars = self.arrays_helper(
                    [statement]
                )
                # We make sure that we only add code to `lines` after it went
                # through completely
                ufunc_lines = []
                # No need to load a variable if it is only in read because of
                # the in-place operation
                if (
                    statement.inplace
                    and variable_indices[statement.var] != "_idx"
                    and statement.var not in get_identifiers(statement.expr)
                ):
                    read = read - {statement.var}
                ufunc_lines.extend(
                    self.read_arrays(read, write, indices, variables, variable_indices)
                )
                ufunc_lines.append(
                    self.ufunc_at_vectorisation(
                        statement,
                        variables,
                        variable_indices,
                        conditional_write_vars,
                        created_vars,
                        used_variables,
                    )
                )
                # Do not write back such values, the ufuncs have modified the
                # underlying array already
                if statement.inplace and variable_indices[statement.var] != "_idx":
                    write = write - {statement.var}
                ufunc_lines.extend(
                    self.write_arrays(
                        [statement], read, write, variables, variable_indices
                    )
                )
                lines.extend(ufunc_lines)
        except VectorisationError:
            if self._use_ufunc_at_vectorisation:
                logger.info(
                    "Failed to vectorise code, falling back on Python loop: note that"
                    " this will be very slow! Switch to another code generation target"
                    " for best performance (e.g. cython). First line is: "
                    + str(statements[0]),
                    once=True,
                )
            lines = []
            lines.extend(
                [
                    "_full_idx = _idx",
                    "for _idx in _full_idx:",
                    "    _vectorisation_idx = _idx",
                ]
            )
            read, write, indices, conditional_write_vars = self.arrays_helper(
                statements
            )
            lines.extend(
                indent(code)
                for code in self.read_arrays(
                    read, write, indices, variables, variable_indices
                )
            )
            for statement in statements:
                line = self.translate_statement(statement)
                if statement.var in conditional_write_vars:
                    lines.append(indent(f"if {conditional_write_vars[statement.var]}:"))
                    lines.append(indent(line, 2))
                else:
                    lines.append(indent(line))
            lines.extend(
                indent(code)
                for code in self.write_arrays(
                    statements, read, write, variables, variable_indices
                )
            )
        return lines

    def read_arrays(self, read, write, indices, variables, variable_indices):
        # index and read arrays (index arrays first)
        lines = []
        for varname in itertools.chain(indices, read):
            var = variables[varname]
            index = variable_indices[varname]
            # if index in iterate_all:
            #                line = '{varname} = {array_name}'
            #            else:
            #                line = '{varname} = {array_name}.take({index})'
            #            line = line.format(varname=varname, array_name=self.get_array_name(var), index=index)
            line = f"{varname} = {self.get_array_name(var)}"
            if index not in self.iterate_all:
                line += f"[{index}]"
            elif varname in write:
                # avoid potential issues with aliased variables, see github #259
                line += ".copy()"
            lines.append(line)
        return lines

    def write_arrays(self, statements, read, write, variables, variable_indices):
        # write arrays
        lines = []
        for varname in write:
            var = variables[varname]
            index_var = variable_indices[varname]
            # check if all operations were inplace and we're operating on the
            # whole vector, if so we don't need to write the array back
            if index_var not in self.iterate_all or varname in read:
                all_inplace = False
            else:
                all_inplace = True
                for stmt in statements:
                    if stmt.var == varname and not stmt.inplace:
                        all_inplace = False
                        break
            if not all_inplace:
                line = self.get_array_name(var)
                if index_var in self.iterate_all:
                    line = f"{line}[:]"
                else:
                    line = f"{line}[{index_var}]"
                line = f"{line} = {varname}"
                lines.append(line)
        return lines

    def conditional_write(
        self, line, stmt, variables, conditional_write_vars, created_vars
    ):
        if stmt.var in conditional_write_vars:
            subs = {}
            index = conditional_write_vars[stmt.var]
            # we replace all var with var[index], but actually we use this repl_string first because
            # we don't want to end up with lines like x[not_refractory[not_refractory]] when
            # multiple substitution passes are invoked
            repl_string = (  # this string shouldn't occur anywhere I hope! :)
                "#$(@#&$@$*U#@)$@(#"
            )
            for varname, var in list(variables.items()):
                if isinstance(var, ArrayVariable) and not var.scalar:
                    subs[varname] = f"{varname}[{repl_string}]"
            # all newly created vars are arrays and will need indexing
            for varname in created_vars:
                subs[varname] = f"{varname}[{repl_string}]"
            # Also index _vectorisation_idx so that e.g. rand() works correctly
            subs["_vectorisation_idx"] = f"_vectorisation_idx[{repl_string}]"

            line = word_substitute(line, subs)
            line = line.replace(repl_string, index)
        return line

    def translate_one_statement_sequence(self, statements, scalar=False):
        variables = self.variables
        variable_indices = self.variable_indices
        read, write, indices, conditional_write_vars = self.arrays_helper(statements)
        lines = []

        all_unique = not self.has_repeated_indices(statements)

        if scalar or all_unique:
            # Simple translation
            lines.extend(
                self.read_arrays(read, write, indices, variables, variable_indices)
            )
            created_vars = {stmt.var for stmt in statements if stmt.op == ":="}
            for stmt in statements:
                line = self.translate_statement(stmt)
                line = self.conditional_write(
                    line, stmt, variables, conditional_write_vars, created_vars
                )
                lines.append(line)
            lines.extend(
                self.write_arrays(statements, read, write, variables, variable_indices)
            )
        else:
            # More complex translation to deal with repeated indices
            lines.extend(self.vectorise_code(statements, variables, variable_indices))

        return lines


################################################################################
# Implement functions
################################################################################
# Functions that exist under the same name in numpy
for func_name, func in [
    ("sin", np.sin),
    ("cos", np.cos),
    ("tan", np.tan),
    ("sinh", np.sinh),
    ("cosh", np.cosh),
    ("tanh", np.tanh),
    ("exp", np.exp),
    ("log", np.log),
    ("log10", np.log10),
    ("sqrt", np.sqrt),
    ("arcsin", np.arcsin),
    ("arccos", np.arccos),
    ("arctan", np.arctan),
    ("abs", np.abs),
    ("sign", np.sign),
]:
    DEFAULT_FUNCTIONS[func_name].implementations.add_implementation(
        NumpyCodeGenerator, code=func
    )


# Functions that are implemented in a somewhat special way
def randn_func(vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.randn(N)
    except TypeError:
        # scalar value
        return np.random.randn()


def rand_func(vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.rand(N)
    except TypeError:
        # scalar value
        return np.random.rand()


def poisson_func(lam, vectorisation_idx):
    try:
        N = len(vectorisation_idx)
        return np.random.poisson(lam, size=N)
    except TypeError:
        # scalar value
        return np.random.poisson(lam)


DEFAULT_FUNCTIONS["randn"].implementations.add_implementation(
    NumpyCodeGenerator, code=randn_func
)
DEFAULT_FUNCTIONS["rand"].implementations.add_implementation(
    NumpyCodeGenerator, code=rand_func
)
DEFAULT_FUNCTIONS["poisson"].implementations.add_implementation(
    NumpyCodeGenerator, code=poisson_func
)
clip_func = lambda array, a_min, a_max: np.clip(array, a_min, a_max)
DEFAULT_FUNCTIONS["clip"].implementations.add_implementation(
    NumpyCodeGenerator, code=clip_func
)
int_func = lambda value: np.int32(value)
DEFAULT_FUNCTIONS["int"].implementations.add_implementation(
    NumpyCodeGenerator, code=int_func
)
ceil_func = lambda value: np.int32(np.ceil(value))
DEFAULT_FUNCTIONS["ceil"].implementations.add_implementation(
    NumpyCodeGenerator, code=ceil_func
)
floor_func = lambda value: np.int32(np.floor(value))
DEFAULT_FUNCTIONS["floor"].implementations.add_implementation(
    NumpyCodeGenerator, code=floor_func
)

# We need to explicitly add an implementation for the timestep function,
# otherwise Brian would *add* units during simulation, thinking that the
# timestep function would not work correctly otherwise. This would slow the
# function down significantly.
DEFAULT_FUNCTIONS["timestep"].implementations.add_implementation(
    NumpyCodeGenerator, code=timestep
)



import ast
from collections import namedtuple

from brian2.utils.stringtools import deindent

__all__ = ["abstract_code_dependencies"]


def get_read_write_funcs(parsed_code):
    allids = set()
    read = set()
    write = set()
    funcs = set()
    for node in ast.walk(parsed_code):
        if node.__class__ is ast.Name:
            allids.add(node.id)
            if node.ctx.__class__ is ast.Store:
                write.add(node.id)
            elif node.ctx.__class__ is ast.Load:
                read.add(node.id)
            else:
                raise SyntaxError
        elif node.__class__ is ast.Call:
            funcs.add(node.func.id)

    read = read - funcs

    # check that there's no funky stuff going on with functions
    if funcs.intersection(write):
        raise SyntaxError("Cannot assign to functions in abstract code")

    return allids, read, write, funcs


def abstract_code_dependencies(code, known_vars=None, known_funcs=None):
    """
    Analyses identifiers used in abstract code blocks

    Parameters
    ----------

    code : str
        The abstract code block.
    known_vars : set
        The set of known variable names.
    known_funcs : set
        The set of known function names.

    Returns
    -------

    results : namedtuple with the following fields
        ``all``
            The set of all identifiers that appear in this code block,
            including functions.
        ``read``
            The set of values that are read, excluding functions.
        ``write``
            The set of all values that are written to.
        ``funcs``
            The set of all function names.
        ``known_all``
            The set of all identifiers that appear in this code block and
            are known.
        ``known_read``
            The set of known values that are read, excluding functions.
        ``known_write``
            The set of known values that are written to.
        ``known_funcs``
            The set of known functions that are used.
        ``unknown_read``
            The set of all unknown variables whose values are read. Equal
            to ``read-known_vars``.
        ``unknown_write``
            The set of all unknown variables written to. Equal to
            ``write-known_vars``.
        ``unknown_funcs``
            The set of all unknown function names, equal to
            ``funcs-known_funcs``.
        ``undefined_read``
            The set of all unknown variables whose values are read before they
            are written to. If this set is nonempty it usually indicates an
            error, since a variable that is read should either have been
            defined in the code block (in which case it will appear in
            ``newly_defined``) or already be known.
        ``newly_defined``
            The set of all variable names which are newly defined in this
            abstract code block.
    """
    if known_vars is None:
        known_vars = set()
    if known_funcs is None:
        known_funcs = set()
    if not isinstance(known_vars, set):
        known_vars = set(known_vars)
    if not isinstance(known_funcs, set):
        known_funcs = set(known_funcs)

    code = deindent(code, docstring=True)
    parsed_code = ast.parse(code, mode="exec")

    # Get the list of all variables that are read from and written to,
    # ignoring the order
    allids, read, write, funcs = get_read_write_funcs(parsed_code)

    # Now check if there are any values that are unknown and read before
    # they are written to
    defined = known_vars.copy()
    newly_defined = set()
    undefined_read = set()
    for line in parsed_code.body:
        _, cur_read, cur_write, _ = get_read_write_funcs(line)
        undef = cur_read - defined
        undefined_read |= undef
        newly_defined |= (cur_write - defined) - undefined_read
        defined |= cur_write

    # Return the results as a named tuple
    results = dict(
        all=allids,
        read=read,
        write=write,
        funcs=funcs,
        known_all=allids.intersection(known_vars.union(known_funcs)),
        known_read=read.intersection(known_vars),
        known_write=write.intersection(known_vars),
        known_funcs=funcs.intersection(known_funcs),
        unknown_read=read - known_vars,
        unknown_write=write - known_vars,
        unknown_funcs=funcs - known_funcs,
        undefined_read=undefined_read,
        newly_defined=newly_defined,
    )
    return namedtuple("AbstractCodeDependencies", list(results.keys()))(**results)



"""
Parameters for spike initiation simulations.
"""
from brian2.units import *

# Passive parameters
EL = -75*mV
S = 7.85e-9*meter**2  # area (sphere of 50 um diameter)
Cm = 0.75*uF/cm**2
gL = 1. / (30000*ohm*cm**2)
Ri = 150*ohm*cm

# Na channels
ENa = 60*mV
ka = 6*mV
va = -40*mV
gNa_0 = gL * 2*S
taum = 0.1*ms



#!/usr/bin/env python
"""
This code contains an adapted version of the voltage-dependent triplet STDP rule from:
Clopath et al., Connectivity reflects coding: a model of voltage-based STDP with homeostasis, Nature Neuroscience, 2010
(http://dx.doi.org/10.1038/nn.2479)

The plasticity rule is adapted for a leaky integrate & fire model in Brian2.
More specifically, the filters ``v_lowpass1`` and ``v_lowpass2`` are
incremented by a constant at every post-synaptic spike time, to
compensate for the lack of an actual spike in the integrate & fire model.

As an illustration of the rule, we simulate the competition between
inputs projecting on a downstream neuron. We would like to note that the
parameters have been chosen arbitrarily to qualitatively reproduce the
behavior of the original work, but need additional fitting.

We kindly ask to cite the article when using the model presented below.

This code was written by Jacopo Bono, 12/2015
"""
from brian2 import *

################################################################################
# PLASTICITY MODEL
################################################################################

#### Plasticity Parameters

V_rest = -70.*mV        # resting potential
V_thresh = -55.*mV      # spiking threshold
Theta_low = V_rest      # depolarization threshold for plasticity
x_reset = 1.            # spike trace reset value
taux = 15.*ms           # spike trace time constant
A_LTD = 1.5e-4          # depression amplitude
A_LTP = 1.5e-2          # potentiation amplitude
tau_lowpass1 = 40*ms    # timeconstant for low-pass filtered voltage
tau_lowpass2 = 30*ms    # timeconstant for low-pass filtered voltage
tau_homeo = 1000*ms     # homeostatic timeconstant
v_target = 12*mV**2     # target depolarisation

#### Plasticity Equations

# equations executed at every timestepC
Syn_model =   ('''
            w_ampa:1                # synaptic weight (ampa synapse)
            ''')

# equations executed only when a presynaptic spike occurs
Pre_eq = ('''
            g_ampa_post += w_ampa*ampa_max_cond                                                               # increment synaptic conductance
            A_LTD_u = A_LTD*(v_homeo**2/v_target)                                                             # metaplasticity
            w_minus = A_LTD_u*(v_lowpass1_post/mV - Theta_low/mV)*int(v_lowpass1_post/mV - Theta_low/mV > 0)  # synaptic depression
            w_ampa = clip(w_ampa-w_minus, 0, w_max)                                                           # hard bounds
            ''' )

# equations executed only when a postsynaptic spike occurs
Post_eq = ('''
            v_lowpass1 += 10*mV                                                                                        # mimics the depolarisation effect due to a spike
            v_lowpass2 += 10*mV                                                                                        # mimics the depolarisation effect due to a spike
            v_homeo += 0.1*mV                                                                                          # mimics the depolarisation effect due to a spike
            w_plus = A_LTP*x_trace_pre*(v_lowpass2_post/mV - Theta_low/mV)*int(v_lowpass2_post/mV - Theta_low/mV > 0)  # synaptic potentiation
            w_ampa = clip(w_ampa+w_plus, 0, w_max)                                                                     # hard bounds
            ''' )

################################################################################
# I&F Parameters and equations
################################################################################

#### Neuron parameters

gleak = 30.*nS                  # leak conductance
C = 300.*pF                     # membrane capacitance
tau_AMPA = 2.*ms                # AMPA synaptic timeconstant
E_AMPA = 0.*mV                  # reversal potential AMPA

ampa_max_cond = 5.e-8*siemens  # Ampa maximal conductance
w_max = 1.                      # maximal ampa weight

#### Neuron Equations

# We connect 10 presynaptic neurons to 1 downstream neuron

# downstream neuron
eqs_neurons = '''
dv/dt = (gleak*(V_rest-v) + I_ext + I_syn)/C: volt      # voltage
dv_lowpass1/dt = (v-v_lowpass1)/tau_lowpass1 : volt     # low-pass filter of the voltage
dv_lowpass2/dt = (v-v_lowpass2)/tau_lowpass2 : volt     # low-pass filter of the voltage
dv_homeo/dt = (v-V_rest-v_homeo)/tau_homeo : volt       # low-pass filter of the voltage
I_ext : amp                                             # external current
I_syn = g_ampa*(E_AMPA-v): amp                          # synaptic current
dg_ampa/dt = -g_ampa/tau_AMPA : siemens                 # synaptic conductance
dx_trace/dt = -x_trace/taux :1                          # spike trace
'''

# input neurons
eqs_inputs = '''
dv/dt = gleak*(V_rest-v)/C: volt                        # voltage
dx_trace/dt = -x_trace/taux :1                          # spike trace
rates : Hz                                              # input rates
selected_index : integer (shared)                       # active neuron
'''

################################################################################
# Simulation
################################################################################

#### Parameters

defaultclock.dt = 500.*us                        # timestep
Nr_neurons = 1                                   # Number of downstream neurons
Nr_inputs = 5                                    # Number of input neurons
input_rate = 35*Hz                               # Rates
init_weight = 0.5                                # initial synaptic weight
final_t = 20.*second                             # end of simulation
input_time = 100.*ms                             # duration of an input

#### Create neuron objects

Nrn_downstream = NeuronGroup(Nr_neurons, eqs_neurons, threshold='v>V_thresh',
                             reset='v=V_rest;x_trace+=x_reset/(taux/ms)',
                             method='euler')
Nrns_input = NeuronGroup(Nr_inputs, eqs_inputs, threshold='rand()<rates*dt',
                         reset='v=V_rest;x_trace+=x_reset/(taux/ms)',
                         method='exact')

#### create Synapses

Syn = Synapses(Nrns_input, Nrn_downstream,
               model=Syn_model,
               on_pre=Pre_eq,
               on_post=Post_eq
               )

Syn.connect(i=numpy.arange(Nr_inputs), j=0)

#### Monitors and storage
W_evolution = StateMonitor(Syn, 'w_ampa', record=True)

#### Run

# Initial values
Nrn_downstream.v = V_rest
Nrn_downstream.v_lowpass1 = V_rest
Nrn_downstream.v_lowpass2 = V_rest
Nrn_downstream.v_homeo = 0
Nrn_downstream.I_ext = 0.*amp
Nrn_downstream.x_trace = 0.
Nrns_input.v = V_rest
Nrns_input.x_trace = 0.
Syn.w_ampa = init_weight

# Switch on a different input every 100ms
Nrns_input.run_regularly('''
                         selected_index = int(floor(rand()*Nr_inputs))
                         rates = input_rate * int(selected_index == i)  # All rates are zero except for the selected neuron
                         ''', dt=input_time)
run(final_t, report='text')

################################################################################
# Plots
################################################################################
stitle = 'Synaptic Competition'

fig = figure(figsize=(8, 5))
for kk in range(Nr_inputs):
    plt.plot(W_evolution.t, W_evolution.w_ampa[kk], '-', linewidth=2)
xlabel('Time [ms]', fontsize=22)
ylabel('Weight [a.u.]', fontsize=22)
plt.subplots_adjust(bottom=0.2, left=0.15, right=0.95, top=0.85)
title(stitle, fontsize=22)
plt.show()



"""
Brian's logging module.

Preferences
-----------
.. document_brian_prefs:: logging
"""

import atexit
import logging
import logging.handlers
import os
import shutil
import sys
import tempfile
import time
from logging.handlers import RotatingFileHandler
from warnings import warn

import numpy

try:
    import scipy
except ImportError:
    scipy = None
import sympy

import brian2
from brian2.core.preferences import BrianPreference, prefs

from .environment import running_from_ipython

__all__ = ["get_logger", "BrianLogger", "std_silent"]

# ===============================================================================
# Logging preferences
# ===============================================================================


def log_level_validator(log_level):
    log_levels = ("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "DIAGNOSTIC")
    return log_level.upper() in log_levels


#: Our new log level for more detailed debug output (mostly useful for debugging
#: Brian itself, not for user scripts)
DIAGNOSTIC = 5

#: Translation from string representation to number
LOG_LEVELS = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "DIAGNOSTIC": DIAGNOSTIC,
}
logging.addLevelName(DIAGNOSTIC, "DIAGNOSTIC")

if "logging" not in prefs.pref_register:
    # Duplicate import of this module can happen when the documentation is built
    prefs.register_preferences(
        "logging",
        "Logging system preferences",
        delete_log_on_exit=BrianPreference(
            default=True,
            docs="""
            Whether to delete the log and script file on exit.

            If set to ``True`` (the default), log files (and the copy of the main
            script) will be deleted after the brian process has exited, unless an
            uncaught exception occurred. If set to ``False``, all log files will be
            kept.
            """,
        ),
        file_log_level=BrianPreference(
            default="DEBUG",
            docs="""
            What log level to use for the log written to the log file.

            In case file logging is activated (see `logging.file_log`), which log
            level should be used for logging. Has to be one of CRITICAL, ERROR,
            WARNING, INFO, DEBUG or DIAGNOSTIC.
            """,
            validator=log_level_validator,
        ),
        console_log_level=BrianPreference(
            default="INFO",
            docs="""
            What log level to use for the log written to the console.

            Has to be one of CRITICAL, ERROR, WARNING, INFO, DEBUG or DIAGNOSTIC.
            """,
            validator=log_level_validator,
        ),
        file_log=BrianPreference(
            default=True,
            docs="""
            Whether to log to a file or not.

            If set to ``True`` (the default), logging information will be written
            to a file. The log level can be set via the `logging.file_log_level`
            preference.
            """,
        ),
        file_log_max_size=BrianPreference(
            default=10000000,
            docs="""
            The maximum size for the debug log before it will be rotated.

            If set to any value ``> 0``, the debug log will be rotated once
            this size is reached. Rotating the log means that the old debug log
            will be moved into a file in the same directory but with suffix ``".1"``
            and the a new log file will be created with the same pathname as the
            original file. Only one backup is kept; if a file with suffix ``".1"``
            already exists when rotating, it will be overwritten.
            If set to ``0``, no log rotation will be applied.
            The default setting rotates the log file after 10MB.
            """,
        ),
        save_script=BrianPreference(
            default=True,
            docs="""
            Whether to save a copy of the script that is run.

            If set to ``True`` (the default), a copy of the currently run script
            is saved to a temporary location. It is deleted after a successful
            run (unless `logging.delete_log_on_exit` is ``False``) but is kept after
            an uncaught exception occured. This can be helpful for debugging,
            in particular when several simulations are running in parallel.
            """,
        ),
        std_redirection=BrianPreference(
            default=True,
            docs="""
            Whether or not to redirect stdout/stderr to null at certain places.

            This silences a lot of annoying compiler output, but will also hide
            error messages making it harder to debug problems. You can always
            temporarily switch it off when debugging. If
            `logging.std_redirection_to_file` is set to ``True`` as well, then the
            output is saved to a file and if an error occurs the name of this file
            will be printed.
            """,
        ),
        std_redirection_to_file=BrianPreference(
            default=True,
            docs="""
            Whether to redirect stdout/stderr to a file.

            If both ``logging.std_redirection`` and this preference are set to
            ``True``, all standard output/error (most importantly output from
            the compiler) will be stored in files and if an error occurs the name
            of this file will be printed. If `logging.std_redirection` is ``True``
            and this preference is ``False``, then all standard output/error will
            be completely suppressed, i.e. neither be displayed nor stored in a
            file.

            The value of this preference is ignore if `logging.std_redirection` is
            set to ``False``.
            """,
        ),
        display_brian_error_message=BrianPreference(
            default=True,
            docs="""
            Whether to display a text for uncaught errors, mentioning the location
            of the log file, the mailing list and the github issues.

            Defaults to ``True``.""",
        ),
    )

# ===============================================================================
# Initial setup
# ===============================================================================


def _encode(text):
    """Small helper function to encode unicode strings as UTF-8."""
    return text.encode("UTF-8")


UNHANDLED_ERROR_MESSAGE = (
    "Brian 2 encountered an unexpected error. "
    "If you think this is a bug in Brian 2, please report this issue either to the "
    "discourse forum at <http://brian.discourse.group/>, "
    "or to the issue tracker at <https://github.com/brian-team/brian2/issues>."
)


def brian_excepthook(exc_type, exc_obj, exc_tb):
    """
    Display a message mentioning the debug log in case of an uncaught
    exception.
    """
    # Do not catch Ctrl+C
    if exc_type == KeyboardInterrupt:
        return
    logger = logging.getLogger("brian2")
    BrianLogger.exception_occured = True

    if not prefs["logging.display_brian_error_message"]:
        # Put the exception message in the log file, but do not log to the
        # console
        if BrianLogger.console_handler is not None:
            logger.removeHandler(BrianLogger.console_handler)
        logger.exception("An exception occured", exc_info=(exc_type, exc_obj, exc_tb))
        if BrianLogger.console_handler is not None:
            logger.addHandler(BrianLogger.console_handler)
        # Run the default except hook
        return sys.__excepthook__(exc_type, exc_obj, exc_tb)

    message = UNHANDLED_ERROR_MESSAGE
    if BrianLogger.tmp_log is not None:
        message += (
            " Please include this file with debug information in your "
            f"report: {BrianLogger.tmp_log} "
        )
    if BrianLogger.tmp_script is not None:
        message += (
            " Additionally, you can also include a copy "
            "of the script that was run, available "
            f"at: {BrianLogger.tmp_script}"
        )
    if hasattr(std_silent, "dest_fname_stdout"):
        stdout = std_silent.dest_fname_stdout
        stderr = std_silent.dest_fname_stderr
        message += (
            " You can also include a copy of the "
            "redirected std stream outputs, available at "
            f"'{stdout}' and '{stderr}'."
        )

    message += " Thanks!"  # very important :)
    logger.error(message, exc_info=(exc_type, exc_obj, exc_tb))


def clean_up_logging():
    """
    Shutdown the logging system and delete the debug log file if no error
    occured.
    """
    BrianLogger.initialized = False
    logging.shutdown()
    if not BrianLogger.exception_occured and prefs["logging.delete_log_on_exit"]:
        if BrianLogger.tmp_log is not None:
            try:
                os.remove(BrianLogger.tmp_log)
            except OSError as exc:
                warn(f"Could not delete log file: {exc}")
            # Remove log files that have been rotated (currently only one)
            rotated_log = f"{BrianLogger.tmp_log}.1"
            if os.path.exists(rotated_log):
                try:
                    os.remove(rotated_log)
                except OSError as exc:
                    warn(f"Could not delete log file: {exc}")
        if BrianLogger.tmp_script is not None:
            try:
                os.remove(BrianLogger.tmp_script)
            except OSError as exc:
                warn(f"Could not delete copy of script file: {exc}")
        std_silent.close()


atexit.register(clean_up_logging)


class HierarchyFilter:
    """
    A class for suppressing all log messages in a subtree of the name hierarchy.
    Does exactly the opposite as the `logging.Filter` class, which allows
    messages in a certain name hierarchy to *pass*.

    Parameters
    ----------
    name : str
        The name hiearchy to suppress. See `BrianLogger.suppress_hierarchy` for
        details.
    """

    def __init__(self, name):
        self.orig_filter = logging.Filter(name)

    def filter(self, record):
        """
        Filter out all messages in a subtree of the name hierarchy.
        """
        # do the opposite of what the standard filter class would do
        return not self.orig_filter.filter(record)


class NameFilter:
    """
    A class for suppressing log messages ending with a certain name.

    Parameters
    ----------
    name : str
        The name to suppress. See `BrianLogger.suppress_name` for details.
    """

    def __init__(self, name):
        self.name = name

    def filter(self, record):
        """
        Filter out all messages ending with a certain name.
        """
        # The last part of the name
        record_name = record.name.split(".")[-1]
        return self.name != record_name


class RemoveBrian2Filter(logging.Filter):
    """
    A class for removing the ``brian2`` prefix from log messages.
    Will be used for extension packages that use the Brian logging system.
    """

    def filter(self, record):
        assert record.name.startswith("brian2."), record.name
        record.name = record.name[7:]
        return True


class BrianLogger:
    """
    Convenience object for logging. Call `get_logger` to get an instance of
    this class.

    Parameters
    ----------
    name : str
        The name used for logging, normally the name of the module. If the name
        does not start with ``brian2.``, it will be prepended automatically when
        interacting with the `logging` module. This means that from the logging's
        system point of view, it will use the configuration for the logger in the
        ``brian2`` hierachy. However, when displaying the name in the log messages,
        the ``brian2.`` prefix will be removed. This is useful for extension
        modules, which can use the Brian logging system, but will be displayed as
        ``myextension`` instead of ``brian2.myextension``. This is also used in
        Brian's test suite, which only considers log messages starting with
        ``brian2``.
    """

    #: Global flag to know whether the logging system is in a usable state
    initialized = False

    #: Class attribute to remember whether any exception occured
    exception_occured = False

    #: Class attribute for remembering log messages that should only be
    #: displayed once
    _log_messages = set()

    #: The name of the temporary log file (by default deleted after the run if
    #: no exception occurred), if any
    tmp_log = None

    #: The `logging.FileHandler` responsible for logging to the temporary log
    #: file
    file_handler = None

    #: The `logging.StreamHandler` responsible for logging to the console
    console_handler = None

    #: The name of the temporary copy of the main script file (by default
    #: deleted after the run if no exception occurred), if any
    tmp_script = None

    #: The pid of the process that initialized the logger – used to switch off file logging in
    #: multiprocessing contexts
    _pid = None

    def __init__(self, name):
        if not name.startswith("brian2."):
            self.name = "brian2." + name
            self.filter_name = True
        else:
            self.name = name
            self.filter_name = False

    def _log(self, log_level, msg, name_suffix, once):
        """
        Log an entry.

        Parameters
        ----------
        log_level : {'debug', 'info', 'warn', 'error'}
            The level with which to log the message.
        msg : str
            The log message.
        name_suffix : str
            A suffix that will be added to the logger name.
        once : bool
            Whether to suppress identical messages if they are logged again.
        """
        if not BrianLogger.initialized:
            # Prevent logging errors on exit
            return
        name = self.name
        if name_suffix:
            name += f".{name_suffix}"

        # Switch off file logging when using multiprocessing
        if BrianLogger.tmp_log is not None and BrianLogger._pid != os.getpid():
            BrianLogger.tmp_log = None
            logging.getLogger("brian2").removeHandler(BrianLogger.file_handler)
            BrianLogger.file_handler = None

        if once:
            # Check whether this exact message has already been displayed
            log_tuple = (name, log_level, msg)
            if log_tuple in BrianLogger._log_messages:
                return
            else:
                BrianLogger._log_messages.add(log_tuple)

        the_logger = logging.getLogger(name)
        if self.filter_name:
            filter = RemoveBrian2Filter()
            the_logger.addFilter(filter)
        the_logger.log(LOG_LEVELS[log_level], msg)
        if self.filter_name:
            the_logger.removeFilter(filter)

    def diagnostic(self, msg, name_suffix=None, once=False):
        """
        Log a diagnostic message.

        Parameters
        ----------
        msg : str
            The message to log.
        name_suffix : str, optional
            A suffix to add to the name, e.g. a class or function name.
        once : bool, optional
            Whether this message should be logged only once and not repeated
            if sent another time.
        """
        self._log("DIAGNOSTIC", msg, name_suffix, once)

    def debug(self, msg, name_suffix=None, once=False):
        """
        Log a debug message.

        Parameters
        ----------
        msg : str
            The message to log.
        name_suffix : str, optional
            A suffix to add to the name, e.g. a class or function name.
        once : bool, optional
            Whether this message should be logged only once and not repeated
            if sent another time.
        """
        self._log("DEBUG", msg, name_suffix, once)

    def info(self, msg, name_suffix=None, once=False):
        """
        Log an info message.

        Parameters
        ----------
        msg : str
            The message to log.
        name_suffix : str, optional
            A suffix to add to the name, e.g. a class or function name.
        once : bool, optional
            Whether this message should be logged only once and not repeated
            if sent another time.
        """
        self._log("INFO", msg, name_suffix, once)

    def warn(self, msg, name_suffix=None, once=False):
        """
        Log a warn message.

        Parameters
        ----------
        msg : str
            The message to log.
        name_suffix : str, optional
            A suffix to add to the name, e.g. a class or function name.
        once : bool, optional
            Whether this message should be logged only once and not repeated
            if sent another time.
        """
        self._log("WARNING", msg, name_suffix, once)

    warning = warn

    def error(self, msg, name_suffix=None, once=False):
        """
        Log an error message.

        Parameters
        ----------
        msg : str
            The message to log.
        name_suffix : str, optional
            A suffix to add to the name, e.g. a class or function name.
        once : bool, optional
            Whether this message should be logged only once and not repeated
            if sent another time.
        """
        self._log("ERROR", msg, name_suffix, once)

    @staticmethod
    def _suppress(filterobj, filter_log_file):
        """
        Apply a filter object to log messages.

        Parameters
        ----------
        filterobj : `logging.Filter`
            A filter object to apply to log messages.
        filter_log_file : bool
            Whether the filter also applies to log messages in the log file.
        """
        BrianLogger.console_handler.addFilter(filterobj)

        if filter_log_file:
            BrianLogger.file_handler.addFilter(filterobj)

    @staticmethod
    def suppress_hierarchy(name, filter_log_file=False):
        """
        Suppress all log messages in a given hiearchy.

        Parameters
        ----------
        name : str
            Suppress all log messages in the given `name` hierarchy. For
            example, specifying ``'brian2'`` suppresses all messages logged
            by Brian, specifying ``'brian2.codegen'`` suppresses all messages
            generated by the code generation modules.
        filter_log_file : bool, optional
            Whether to suppress the messages also in the log file. Defaults to
            ``False`` meaning that suppressed messages are not displayed on
            the console but are still saved to the log file.
        """

        suppress_filter = HierarchyFilter(name)

        BrianLogger._suppress(suppress_filter, filter_log_file)

    @staticmethod
    def suppress_name(name, filter_log_file=False):
        """
        Suppress all log messages with a given name.

        Parameters
        ----------
        name : str
            Suppress all log messages ending in the given `name`. For
            example, specifying ``'resolution_conflict'`` would suppress
            messages with names such as
            ``brian2.equations.codestrings.CodeString.resolution_conflict`` or
            ``brian2.equations.equations.Equations.resolution_conflict``.
        filter_log_file : bool, optional
            Whether to suppress the messages also in the log file. Defaults to
            ``False`` meaning that suppressed messages are not displayed on
            the console but are still saved to the log file.
        """
        suppress_filter = NameFilter(name)

        BrianLogger._suppress(suppress_filter, filter_log_file)

    @staticmethod
    def log_level_diagnostic():
        """
        Set the log level to "diagnostic".
        """
        BrianLogger.console_handler.setLevel(DIAGNOSTIC)

    @staticmethod
    def log_level_debug():
        """
        Set the log level to "debug".
        """
        BrianLogger.console_handler.setLevel(logging.DEBUG)

    @staticmethod
    def log_level_info():
        """
        Set the log level to "info".
        """
        BrianLogger.console_handler.setLevel(logging.INFO)

    @staticmethod
    def log_level_warn():
        """
        Set the log level to "warn".
        """
        BrianLogger.console_handler.setLevel(logging.WARN)

    @staticmethod
    def log_level_error():
        """
        Set the log level to "error".
        """
        BrianLogger.console_handler.setLevel(logging.ERROR)

    @staticmethod
    def initialize():
        """
        Initialize Brian's logging system. This function will be called
        automatically when Brian is imported.
        """
        # get the main logger
        logger = logging.getLogger("brian2")
        logger.propagate = False
        logger.setLevel(LOG_LEVELS["DIAGNOSTIC"])

        # Log to a file
        if prefs["logging.file_log"]:
            try:
                # Temporary filename used for logging
                with tempfile.NamedTemporaryFile(
                    prefix="brian_debug_", suffix=".log", delete=False
                ) as tmp_f:
                    BrianLogger.tmp_log = tmp_f.name
                # Remove any previously existing file handler
                if BrianLogger.file_handler is not None:
                    BrianLogger.file_handler.close()
                    logger.removeHandler(BrianLogger.file_handler)
                # Rotate log file after prefs['logging.file_log_max_size'] bytes and keep one copy
                BrianLogger.file_handler = RotatingFileHandler(
                    BrianLogger.tmp_log,
                    mode="a",
                    maxBytes=prefs["logging.file_log_max_size"],
                    backupCount=1,
                    encoding="utf-8",
                )
                BrianLogger.file_handler.setLevel(
                    LOG_LEVELS[prefs["logging.file_log_level"].upper()]
                )
                BrianLogger.file_handler.setFormatter(
                    logging.Formatter(
                        "%(asctime)s %(levelname)-10s %(name)s: %(message)s"
                    )
                )
                logger.addHandler(BrianLogger.file_handler)
                BrianLogger._pid = os.getpid()
            except OSError as ex:
                warn(f"Could not create log file: {ex}")

        # Save a copy of the script
        BrianLogger.tmp_script = None
        if prefs["logging.save_script"]:
            if (
                len(sys.argv[0])
                and not running_from_ipython()
                and os.path.isfile(sys.argv[0])
            ):
                try:
                    tmp_file = tempfile.NamedTemporaryFile(
                        prefix="brian_script_", suffix=".py", delete=False
                    )
                    with tmp_file:
                        # Timestamp
                        tmp_file.write(_encode(f"# {time.asctime()}\n"))
                        # Command line arguments
                        tmp_file.write(_encode(f"# Run as: {' '.join(sys.argv)}\n\n"))
                        # The actual script file
                        # TODO: We are copying the script file as it is, this might clash
                        # with the encoding we used for the comments added above
                        with open(os.path.abspath(sys.argv[0]), "rb") as script_file:
                            shutil.copyfileobj(script_file, tmp_file)
                        BrianLogger.tmp_script = tmp_file.name
                except OSError as ex:
                    warn(f"Could not copy script file to temp directory: {ex}")

        if BrianLogger.console_handler is not None:
            logger.removeHandler(BrianLogger.console_handler)

        # create console handler with a higher log level
        BrianLogger.console_handler = logging.StreamHandler()
        BrianLogger.console_handler.setLevel(
            LOG_LEVELS[prefs["logging.console_log_level"]]
        )
        BrianLogger.console_handler.setFormatter(
            logging.Formatter("%(levelname)-10s %(message)s [%(name)s]")
        )

        # add the handler to the logger
        logger.addHandler(BrianLogger.console_handler)

        # We want to log all warnings
        logging.captureWarnings(True)  # pylint: disable=E1101
        # Manually connect to the warnings logger so that the warnings end up in
        # the log file. Note that connecting to the console handler here means
        # duplicated warning messages in the ipython notebook, but not doing so
        # would mean that they are not displayed at all in the standard ipython
        # interface...
        warn_logger = logging.getLogger("py.warnings")
        warn_logger.addHandler(BrianLogger.console_handler)

        # Put some standard info into the log file
        logger.log(
            logging.DEBUG,
            f"Logging to file: {BrianLogger.tmp_log}, copy of main script saved as:"
            f" {BrianLogger.tmp_script}",
        )
        logger.log(logging.DEBUG, f"Python interpreter: {sys.executable}")
        logger.log(logging.DEBUG, f"Platform: {sys.platform}")
        version_infos = {
            "brian": brian2.__version__,
            "numpy": numpy.__version__,
            "scipy": scipy.__version__ if scipy else "not installed",
            "sympy": sympy.__version__,
            "python": sys.version,
        }
        for _name, _version in version_infos.items():
            logger.log(logging.DEBUG, f"{_name} version is: {str(_version)}")
        # Handle uncaught exceptions
        sys.excepthook = brian_excepthook

        BrianLogger.initialized = True


def get_logger(module_name="brian2"):
    """
    Get an object that can be used for logging.

    Parameters
    ----------
    module_name : str
        The name used for logging, should normally be the module name as
        returned by ``__name__``.

    Returns
    -------
    logger : `BrianLogger`
    """

    return BrianLogger(module_name)


class catch_logs:
    """
    A context manager for catching log messages. Use this for testing the
    messages that are logged. Defaults to catching warning/error messages and
    this is probably the only real use case for testing. Note that while this
    context manager is active, *all* log messages are suppressed. Using this
    context manager returns a list of (log level, name, message) tuples.

    Parameters
    ----------
    log_level : int or str, optional
        The log level above which messages are caught.
    only_from : list, optional
        A list of module names from which messages are caught. Defaults to
        the ``brian2`` module.
    Examples
    --------
    >>> logger = get_logger('brian2.logtest')
    >>> logger.warn('An uncaught warning') # doctest: +SKIP
    WARNING  brian2.logtest: An uncaught warning
    >>> with catch_logs() as l:
    ...    logger.warn('a caught warning')
    ...    print('l contains: %s' % l)
    ...
    l contains: [('WARNING', 'brian2.logtest', 'a caught warning')]

    """

    _entered = False

    def __init__(self, log_level=logging.WARN, only_from=("brian2",)):
        self.log_list = []
        self.handler = LogCapture(self.log_list, log_level, only_from=only_from)
        self._entered = False

    def __enter__(self):
        if self._entered:
            raise RuntimeError(f"Cannot enter {self!r} twice")
        self._entered = True
        return self.log_list

    def __exit__(self, *exc_info):
        if not self._entered:
            raise RuntimeError(f"Cannot exit {self!r} without entering first")
        self.handler.uninstall()


class LogCapture(logging.Handler):
    """
    A class for capturing log warnings. This class is used by
    `~brian2.utils.logger.catch_logs` to allow testing in a similar
    way as with `warnings.catch_warnings`.
    """

    def __init__(self, log_list, log_level=logging.WARN, only_from=("brian2",)):
        logging.Handler.__init__(self, level=log_level)
        self.log_list = log_list
        self.only_from = only_from
        # make a copy of the previous handlers
        self.handlers = list(logging.getLogger("brian2").handlers)
        self.install()

    def emit(self, record):
        # Append a tuple consisting of (level, name, msg) to the list of
        # log messages
        if any(
            record.name == name or record.name.startswith(name + ".")
            for name in self.only_from
        ):
            self.log_list.append((record.levelname, record.name, record.msg))

    def install(self):
        """
        Install this handler to catch all warnings. Temporarily disconnect all
        other handlers.
        """
        the_logger = logging.getLogger("brian2")
        for handler in self.handlers:
            the_logger.removeHandler(handler)
        the_logger.addHandler(self)

    def uninstall(self):
        """
        Uninstall this handler and re-connect the previously installed
        handlers.
        """
        the_logger = logging.getLogger("brian2")
        for handler in self.handlers:
            the_logger.addHandler(handler)


# See http://stackoverflow.com/questions/26126160/redirecting-standard-out-in-err-back-after-os-dup2
# for an explanation of how this function works. Note that 1 and 2 are the file
# numbers for stdout and stderr
class std_silent:
    """
    Context manager that temporarily silences stdout and stderr but keeps the
    output saved in a temporary file and writes it if an exception is raised.
    """

    dest_stdout = None
    dest_stderr = None

    def __init__(self, alwaysprint=False):
        self.alwaysprint = alwaysprint or not prefs["logging.std_redirection"]
        self.redirect_to_file = prefs["logging.std_redirection_to_file"]
        if (
            not self.alwaysprint
            and self.redirect_to_file
            and std_silent.dest_stdout is None
        ):
            std_silent.dest_fname_stdout = tempfile.NamedTemporaryFile(
                prefix="brian_stdout_", suffix=".log", delete=False
            ).name
            std_silent.dest_fname_stderr = tempfile.NamedTemporaryFile(
                prefix="brian_stderr_", suffix=".log", delete=False
            ).name
            std_silent.dest_stdout = open(std_silent.dest_fname_stdout, "w")
            std_silent.dest_stderr = open(std_silent.dest_fname_stderr, "w")

    def __enter__(self):
        if not self.alwaysprint and self.redirect_to_file:
            sys.stdout.flush()
            sys.stderr.flush()
            self.orig_out_fd = os.dup(1)
            self.orig_err_fd = os.dup(2)
            os.dup2(std_silent.dest_stdout.fileno(), 1)
            os.dup2(std_silent.dest_stderr.fileno(), 2)

    def __exit__(self, exc_type, exc_value, traceback):
        if not self.alwaysprint and self.redirect_to_file:
            std_silent.dest_stdout.flush()
            std_silent.dest_stderr.flush()
            if exc_type is not None:
                with open(std_silent.dest_fname_stdout) as f:
                    out = f.read()
                with open(std_silent.dest_fname_stderr) as f:
                    err = f.read()
            os.dup2(self.orig_out_fd, 1)
            os.dup2(self.orig_err_fd, 2)
            os.close(self.orig_out_fd)
            os.close(self.orig_err_fd)
            if exc_type is not None:
                sys.stdout.write(out)
                sys.stderr.write(err)

    @classmethod
    def close(cls):
        if std_silent.dest_stdout is not None:
            std_silent.dest_stdout.close()
            if prefs["logging.delete_log_on_exit"]:
                try:
                    os.remove(std_silent.dest_fname_stdout)
                except OSError:
                    # TODO: this happens quite frequently - why?
                    # The file objects are closed as far as Python is concerned,
                    # but maybe Windows is still hanging on to them?
                    pass
        if std_silent.dest_stderr is not None:
            std_silent.dest_stderr.close()
            if prefs["logging.delete_log_on_exit"]:
                try:
                    os.remove(std_silent.dest_fname_stderr)
                except OSError:
                    pass



#!/usr/bin/env python3
"""
Fig. 2 from:

Real-Time Computing Without Stable States: A New
Framework for Neural Computation Based on Perturbations

Neural Computation 14, 2531–2560 (2002)

by Maass W., Natschläger T. and Markram H.

Sebastian Schmitt, 2022
"""
from collections import defaultdict
import multiprocessing

import numpy as np
import matplotlib.pyplot as plt

from brian2 import (
    NeuronGroup,
    Synapses,
    SpikeGeneratorGroup,
    SpikeMonitor,
    Network,
    prefs,
)
from brian2 import ms, mV, Mohm, nA, second, Hz
from brian2 import defaultclock, prefs

N_NEURONS = 135
V_THRESH = 15 * mV
V_RESET = 13.5 * mV

STIMULUS_POISSON_RATE = 20 * Hz
TARGET_DISTANCES = [0.4, 0.2, 0.1]
N_PAIRS = 200

DT = 0.1 * ms
DURATION = 500 * ms
TS = np.arange(0, DURATION / ms, DT / ms)


def exponential_convolution(t, spikes, tau):
    """Convolute spikes with exponential kernel
    t -- numpy array of times to evaluate the convolution
    spikes -- iterable of spike times
    tau -- exponential decay constant
    """
    if len(spikes):
        return sum([np.exp(-((t - st) / tau)) * (t >= st) for st in spikes])
    else:
        return np.zeros(len(TS))


def gaussian_convolution(t, spikes, tau):
    """Convolute spikes with Gaussian kernel
    t -- numpy array of times to evaluate the convolution
    spikes -- iterable of spike times
    tau -- exponential decay constant
    """
    if len(spikes):
        return sum([np.exp(-(((t - st) / tau) ** 2)) for st in spikes])
    else:
        return np.zeros(len(TS))


def euclidian_distance(liquid_states_u, liquid_states_v):
    """Euclidian distance between liquid states
    liquid_states_u -- liquid states
    liquid_states_v -- other liquid states

    To match the numbers in the paper, the square root is omitted
    """

    return np.mean((liquid_states_u - liquid_states_v) ** 2, axis=0)


def distance(conv_a, conv_b, dt):
    """Difference of convolutions in the L2-norm
    conv_a -- convolutions
    conv_b -- other convolutions
    dt -- time step

    To match the numbers in the paper, the square root is omitted
    """

    return sum((conv_a - conv_b) ** 2) * dt


def generate_poisson(duration, rate):
    """Generate Poisson spike train
    duration -- duration of spike train
    rate -- rate of spike train

    Return only spike trains that do not have multiple spikes per time bin
    """
    while True:
        N = np.random.poisson(rate * duration)
        spikes = np.random.uniform(0, duration, N)

        spikes_orig = np.sort(spikes)
        shift = 1e-3 * (DT / ms)
        timebins = ((spikes_orig + shift) / (DT / ms)).astype(np.int32)

        if not any(np.diff(timebins) == 0):
            return spikes_orig


def collect_stimulus_pairs():
    """Collect pairs of input stimuli close in target distance"""
    DELTA_DISTANCE = 0.01
    collected_pairs = defaultdict(list)

    while True:

        spikes_u = generate_poisson(DURATION / ms, STIMULUS_POISSON_RATE / Hz / 1e3)
        spikes_v = generate_poisson(DURATION / ms, STIMULUS_POISSON_RATE / Hz / 1e3)

        conv_u = gaussian_convolution(TS, spikes_u, tau=5)
        conv_v = gaussian_convolution(TS, spikes_v, tau=5)

        normed_distance = distance(conv_u, conv_v, DT / ms) / (DURATION / ms)

        for target_distance in TARGET_DISTANCES:
            if (
                abs(normed_distance - target_distance) < DELTA_DISTANCE
                and len(collected_pairs[target_distance]) < N_PAIRS
            ):
                collected_pairs[target_distance].append((spikes_u, spikes_v))

        # stop if we have enough pairs collected
        if len(collected_pairs) == len(TARGET_DISTANCES) and all(
            np.array(list(map(len, collected_pairs.values()))) == N_PAIRS
        ):
            break

    return collected_pairs


def get_neurons():
    neurons = NeuronGroup(
        N_NEURONS,
        """
        tau_mem : second (shared, constant)
        tau_refrac : second (constant)
        v_reset : volt (shared, constant)
        v_thresh : volt (shared, constant)
        I_b : ampere (shared, constant)
        tau_stimulus : second (constant)
        I_syn_ee_synapses : ampere
        I_syn_ei_synapses : ampere
        I_syn_ie_synapses : ampere
        I_syn_ii_synapses : ampere
        dI_stimulus/dt = -I_stimulus/tau_stimulus : ampere
        R_in : ohm
        dv/dt = -v/tau_mem + (I_syn_ee_synapses +
                              I_syn_ei_synapses +
                              I_syn_ie_synapses +
                              I_syn_ii_synapses)*R_in/tau_mem
                           + I_b*R_in/tau_mem
                           + I_stimulus*R_in/tau_mem: volt (unless refractory)
        x_pos : 1 (constant)
        y_pos : 1 (constant)
        z_pos : 1 (constant)
        """,
        threshold="v>v_thresh",
        reset="v=v_reset",
        refractory="tau_refrac",
        method="exact",
        name="neurons",
    )

    neurons.tau_mem = 30 * ms
    neurons.v_thresh = V_THRESH
    neurons.v_reset = V_RESET

    neurons.I_b = 13.5 * nA

    neurons.v[:] = (
        np.random.uniform(V_RESET / mV, V_THRESH / mV, size=len(neurons)) * mV
    )

    neurons.R_in = 1 * Mohm

    # to randomly assign excitatory and inhibitory neurons later
    indices = np.arange(len(neurons))
    np.random.shuffle(indices)

    # a column of 15x3x3 neurons
    neurons.x_pos = indices % 3
    neurons.y_pos = (indices // 3) % 3
    neurons.z_pos = indices // 9

    return neurons


def get_synapses(name, source, target, C, l, tau_I, A, U, D, F, delay):
    synapses_eqs = """
    A : ampere (constant)
    U : 1 (constant)
    tau_I : second (shared, constant)
    D : second (constant)
    dx/dt =  z/D       : 1 (clock-driven) # recovered
    dy/dt = -y/tau_I   : 1 (clock-driven) # active
    z = 1 - x - y      : 1                # inactive
    I_syn_{}_post = A*y : ampere (summed)
    """.format(name)

    if F:
        synapses_eqs += """
        du/dt = -u/F : 1 (clock-driven)
        F : second (constant)
        """

        synapses_action = """
        u += U*(1-u)
        y += u*x # important: update y first
        x += -u*x
        """
    else:
        synapses_action = """
        y += U*x # important: update y first
        x += -U*x
        """

    synapses = Synapses(
        source,
        target,
        model=synapses_eqs,
        on_pre=synapses_action,
        method="exact",
        name=name,
        delay=delay,
    )

    synapses.connect(
        p=f"{C} * exp(-((x_pos_pre-x_pos_post)**2 + (y_pos_pre-y_pos_post)**2 + (z_pos_pre-z_pos_post)**2)/{l}**2)"
    )

    N_syn = len(synapses)

    synapses.tau_I = tau_I

    synapses.A[:] = np.sign(A / nA) * np.random.gamma(1, abs(A / nA), size=N_syn) * nA

    synapses.U[:] = np.random.normal(U, 0.5 * U, size=N_syn)
    # paper samples from uniform, we take the mean
    synapses.U[:][synapses.U < 0] = U

    synapses.D[:] = np.random.normal(D / ms, 0.5 * D / ms, size=N_syn) * ms
    # paper samples from uniform, we take the mean
    synapses.D[:][synapses.D / ms <= 0] = D

    # start fully recovered
    synapses.x = 1

    if F:
        synapses.F[:] = np.random.normal(F / ms, 0.5 * F / ms, size=N_syn) * ms
        # paper samples from uniform, we take the mean
        synapses.F[:][synapses.F / ms <= 0] = F

    return synapses


def sim(net, spike_times):
    """Run network with given stimulus

    Redraws initial membrane voltages

    net -- the network to simulate
    spike_times -- the stimulus to inject
    """
    net.restore()

    net["neurons"].v = (
        np.random.uniform(V_RESET / mV, V_THRESH / mV, size=len(neurons)) * mV
    )
    net["stimulus"].set_spikes([0] * len(spike_times), spike_times * ms)

    net.run(DURATION)

    spikes = list(net["spike_monitor_exc"].spike_trains().values()) + list(
        net["spike_monitor_inh"].spike_trains().values()
    )

    liquid_states = np.array(
        [exponential_convolution(TS, st / ms, tau=30) for st in spikes]
    )

    return liquid_states

if __name__ == '__main__':
    neurons = get_neurons()

    N_exc = int(0.8 * len(neurons))

    exc_neurons = neurons[:N_exc]
    exc_neurons.tau_refrac = 3 * ms
    exc_neurons.tau_stimulus = 3 * ms

    inh_neurons = neurons[N_exc:]
    inh_neurons.tau_refrac = 2 * ms
    inh_neurons.tau_stimulus = 6 * ms

    l_lambda = 2

    ee_synapses = get_synapses(
        "ee_synapses",
        exc_neurons,
        exc_neurons,
        C=0.3,
        l=l_lambda,
        tau_I=3 * ms,
        A=30 * nA,
        U=0.5,
        D=1.1 * second,
        F=0.05 * second,
        delay=1.5 * ms,
    )
    ei_synapses = get_synapses(
        "ei_synapses",
        exc_neurons,
        inh_neurons,
        C=0.2,
        l=l_lambda,
        tau_I=3 * ms,
        A=60 * nA,
        U=0.05,
        D=0.125 * second,
        F=1.2 * second,
        delay=0.8 * ms,
    )
    ie_synapses = get_synapses(
        "ie_synapses",
        inh_neurons,
        exc_neurons,
        C=0.4,
        l=l_lambda,
        tau_I=6 * ms,
        A=-19 * nA,
        U=0.25,
        D=0.7 * second,
        F=0.02 * second,
        delay=0.8 * ms,
    )
    ii_synapses = get_synapses(
        "ii_synapses",
        inh_neurons,
        inh_neurons,
        C=0.1,
        l=l_lambda,
        tau_I=6 * ms,
        A=-19 * nA,
        U=0.32,
        D=0.144 * second,
        F=0.06 * second,
        delay=0.8 * ms,
    )

    # place holder for stimulus
    stimulus = SpikeGeneratorGroup(1, [], [] * ms, name="stimulus")

    spike_monitor_stimulus = SpikeMonitor(stimulus)

    static_synapses_exc = Synapses(
        stimulus,
        exc_neurons,
        "A : ampere (shared, constant)",
        on_pre="I_stimulus += A"
    )
    static_synapses_exc.connect(p=1)
    static_synapses_exc.A = 18 * nA

    static_synapses_inh = Synapses(
        stimulus,
        inh_neurons,
        "A : ampere (shared, constant)",
        on_pre="I_stimulus += A"
    )
    static_synapses_inh.connect(p=1)
    static_synapses_inh.A = 9 * nA

    spike_monitor_exc = SpikeMonitor(exc_neurons, name="spike_monitor_exc")
    spike_monitor_inh = SpikeMonitor(inh_neurons, name="spike_monitor_inh")

    defaultclock.dt = DT

    net = Network(
        [
            neurons,
            ee_synapses,
            ei_synapses,
            ie_synapses,
            ii_synapses,
            static_synapses_exc,
            static_synapses_inh,
            stimulus,
            spike_monitor_exc,
            spike_monitor_inh,
        ]
    )
    net.store()

    collected_pairs = collect_stimulus_pairs()

    # add only jittered pairs
    collected_pairs[0] = [
        [generate_poisson(DURATION / ms, STIMULUS_POISSON_RATE / Hz / 1e3)] * 2
        for _ in range(N_PAIRS)
    ]

    def map_sim(spike_times):
        """Wrapper to sim for multiprocessing
        """
        return sim(net, spike_times)

    result = defaultdict(list)
    # loop over all distances and Poisson stimulus pairs
    for d, pairs in collected_pairs.items():

        with multiprocessing.Pool() as p:
            states_u = p.map(map_sim, [p[0] for p in pairs])
            states_v = p.map(map_sim, [p[1] for p in pairs])

        for liquid_states_u, liquid_states_v in zip(states_u, states_v):
            ed = euclidian_distance(liquid_states_u, liquid_states_v)
            result[d].append(ed)
    # plot
    fig, ax = plt.subplots(figsize=(5, 5))

    linestyles = ["dashed", (0, (8, 6, 1, 6)), (0, (5, 10)), "solid"]

    for d, ls in zip(TARGET_DISTANCES + [0], linestyles):

        eds = result[d]
        eds = np.array(eds)

        ax.plot(
            TS / 1000, np.mean(eds, axis=0), label=f"d(u,v)={d}", linestyle=ls, color="k"
        )

    ax.set_xlabel("time [sec]")
    ax.set_ylabel("state distance")

    ax.set_xlim(0, 0.5)
    ax.set_ylim(0, 2.5)

    ax.legend(loc="upper center", fontsize="x-large", frameon=False)

    plt.show()



{% extends 'common_group.pyx' %}

{% block maincode %}
    {# USES_VARIABLES { _spikespace, neuron_index, _timebins, _period_bins, _lastindex, t_in_timesteps, N } #}

    cdef int32_t _the_period    = {{_period_bins}}
    cdef int32_t _timebin       = {{t_in_timesteps}}
    cdef int32_t _cpp_numspikes = 0;

    if _the_period > 0:
        _timebin %= _the_period
        # If there is a periodicity in the SpikeGenerator, we need to reset the
        # lastindex when the period has passed
        if {{_lastindex}} > 0 and {{_timebins}}[{{_lastindex}} - 1] >= _timebin:
            {{_lastindex}} = 0

    for _idx in range({{_lastindex}}, _num{{_timebins}}):
        if ({{_timebins}}[_idx] > _timebin):
            break

        {{_spikespace}}[_cpp_numspikes] = {{neuron_index}}[_idx]
        _cpp_numspikes += 1

    {{_spikespace}}[N] = _cpp_numspikes

    {{_lastindex}} += _cpp_numspikes

{% endblock %}
